subroutine GL_grid_dp(nmax, lat, lon, w)
!
!==========================================================================
!
! DESCRIPTION: This subroutine generates latitudes, longitudes and weights of
!              the grid nodes required by the Gauss--Legendre quadrature
!              (e.g., Sneeuw, 1994) for a specified maximum harmonic degree.
!
!              The routine is written in double precision.
!
!              Note that the grid is non-equiangular in terms of latitudes.
!              The step in the longitudinal direction is, however, constant.
!
!
! INPUTS:   "nmax" -- Maximum spherical harmonic degree of the signal that
!                     is to be harmonically analyzed
!
!
! OUTPUTS:  "lat" -- Vector of spherical latitudes in radians.
!                    Dimension (nmax + 1).
!
!           "lon" -- Vector of longitudes in radians.
!                    Dimmension (2 * nmax + 2).
!
!           "w"   -- Weights given by the Gauss--Legendre quadrature
!                    (dimensionless).
!                    Dimension (nmax + 1).
!
!
! This subroutine is based on the MATLAB "legzo.m" function
! (https://github.com/Pazus/Legendre-Gauss-Quadrature/blob/master/legzo.m)
! pubslished by Pazus (https://github.com/Pazus). Minor modifications were
! introduced, such that it now yields spherical latitudes and longitudes as the 
! output coordinates.
!
!
! REFERENCES: Sneeuw, N. (1994) Global spherical harmonic analysis by
!                least-squares and numerical quadrature methods in historical
!                perspective. Geophysical Journal International 118:707--716
!
! ==========================================================================


    ! Import modules
    ! ===============================================
    use vartypes,  only: dp            ! Module defining number formats
    use constants, only: pi_dp, eps_dp ! Module defining constants (pi, etc.)
    ! ===============================================


    ! Declaration of variables
    ! ===============================================
    implicit none

    integer, intent(in)   :: nmax
    integer               :: nmax1, m, i, j
    real(dp)              :: z, z1, c, p1, p2, p3, pp
    real(dp), intent(out) :: lat(nmax + 1), lon(2 * nmax + 2), w(nmax + 1)
    ! ===============================================


    ! The main computational part of the subroutine
    ! ===============================================
    nmax1 = nmax + 1

    m = (nmax1 + 1) / 2  ! Note that "m" is rounded to the greatest integer
                         ! less than or equal to "(nmax1 + 1) / 2".

    c = (real(nmax1, dp) + 0.5_dp)


    ! Computation of the latitudes
    ! --------------------------------------------------------------------------
    do i = 1, m

        z = cos(pi_dp * (real(i, dp) - 0.25_dp) / c)
        z1 = z + 1.0_dp

        do while (abs(z - z1) > eps_dp)

            p1 = 1.0_dp
            p2 = 0.0_dp

            do j = 1, nmax1

                p3 = p2
                p2 = p1
                p1 = ((2.0_dp * real(j, dp) - 1.0_dp) * z * p2 - &
                     (real(j, dp) - 1.0_dp) * p3) / real(j, dp)

            end do

            pp = real(nmax1, dp) * (z * p1 - p2) / (z * z - 1.0_dp)
            z1 = z
            z  = z1 - p1 / pp

        end do

        lat(nmax1 + 1 - i) = asin(z)
        lat(i) = -lat(nmax1 + 1 - i) 

        w(nmax1 + 1 - i) = 2.0_dp / ((1.0_dp - z * z) * pp * pp)
        w(i) = w(nmax1 + 1 - i)

    end do ! End of the loop over latitudes
    ! --------------------------------------------------------------------------


    ! Computation of the longitudes
    ! --------------------------------------------------------------------------
    do i = 1, (2 * nmax1)
        lon(i) = pi_dp / real(nmax1, dp) * (real(i, dp) - 1.0_dp)
    end do ! End of the loop over longitudes
    ! --------------------------------------------------------------------------
    ! ===============================================

end subroutine GL_grid_dp
