program Test_run_dp
!
! ==========================================================================
!
! This is a Fortran program to test the "Surface_SHS_dp.f95" and
! "Surface_SHA_dp.f95" subroutines designed for ultra-high-degree (tens of
! thousands and beyond) surface spherical harmonic synthesis and analysis,
! respectively. The program is written in double precision and consists of two 
! examples that demonstrate correctness of these subroutines in a closed-loop 
! environment.
!
!
! The loops over the grid latitudes inside the "Surface_SHS_dp.f95" and
! "Surface_SHA_dp.f95" subroutines are paralellized using OpenMP. By default,
! the "Test_run_dp.f95" program employs automatically all available threads
! (the value can be changed, however; see the variable "nmaxthreads").
!
!
! NOTE: In case of parallel computation using OpenMP, the user may need to set
!       the value of the "OMP_STACKSIZE" environment variable to a sufficiently
!       large value as well as to set the OS limits. For instance, for a
!       maximum degree being equal to a few thousands, the following settings 
!       may be sufficient (Linux bash commands):
!
!       ulimit -s unlimited
!       export OMP_STACKSIZE=2g
!
!
! For FFT in the "Surface_SHA_dp.f95" and "Surface_SHA_qp.f95" subroutines, the
! FFTW package is employed (http://www.fftw.org/). It is assumed that FFTW is
! installed on your machine both in double and quadruple precision with OpenMP
! enabled. For info on the installation of FFTW, please visit the website. In
! case you have only the double precision version of FFTW installed and don't
! want to use the "SHA_SHS" package in quadruple precision, please see also the
! "FFTW3.f95" file.
!
! ==========================================================================


    ! Import modules
    ! ===============================================
    use vartypes,  only: dp    ! Module defining number formats
    use constants, only: pi_dp ! Module defining constants (pi, etc.)
    !$ use omp_lib, only: omp_get_max_threads ! Module to get the maximum
                                              ! number of available threads. By
                                              ! default, the maximum number is
                                              ! used for parallel parts of the
                                              ! synthesis and analysis. However,
                                              ! the value can be changed in the
                                              ! code below (look for
                                              ! "omp_set_num_threads").
    ! ===============================================


    ! Declaration of variables
    ! ===============================================
    implicit none

    integer :: nmax = 360 ! Maximum harmonic degree of the synthesis

    integer :: nlat       ! Number of latitudes in one meridian
    integer :: nlon       ! Number of longitudes in one latitude parallel

    real(dp), allocatable :: lat(:) ! Latitudes in one meridian of the
                                    ! computational grid (Examples No. 1 and 2)
    real(dp), allocatable :: lon(:) ! Longitudes in one latitude parallel of the
                                    ! computational grid (Examples No. 1 and 2)

    ! Minumum and maximum latitudes and longitudes for computing the mean
    ! values over grid cells (Example No. 3)
    real(dp), allocatable :: lat1(:), lat2(:), lon1(:), lon2(:)

    real(dp), allocatable :: w(:)   ! Gauss--Legendre weights for latitudes

    real(dp), allocatable :: cs(:, :)  ! Input spherical harmonic coefficients
                                       ! from the attached
                                       ! "Earth2014.RET2014.degree360.txt" file

    real(dp), allocatable :: f(:, :)   ! Signal synthesized from "cs"

    real(dp), allocatable :: cs2(:, :) ! Output spherical harmonic coefficients
                                       ! from spherical harmonic analysis
    real(dp), allocatable :: cs3(:, :) ! Another output spherical harmonic
                                       ! coefficients from spherical harmonic
                                       ! analysis
    real(dp), allocatable :: cs4(:, :) ! Output spherical harmonic
                                       ! coefficients from spherical harmonic
                                       ! analysis of the mean values

    character(len = 256) :: path ! Absolute path to the directory, where the
                                 ! data file "Earth2014.RET2014.degree360.txt"
                                 ! is stored. This variable needs to be set
                                 ! correctly by the user to run this test
                                 ! program (see below).

    integer :: n, fid  ! Loop variable and unit number for file reading,
                       ! respectively

    integer :: nmaxthreads  ! Number of threads used for parallel loops over
                            ! latitudes in the "Surface_SHA.f95" and
                            ! "Surface_SHS.f95" subroutines
    ! ===============================================


    ! Main computational part of the program
    ! ===============================================

    print *, 'Starting the test computation...'


    ! Absolute or relative path to the directory, where the data file
    ! "Earth2014.RET2014.degree360.txt" is stored.
    path = './data'


    ! Get and then set the maximum number of available threads for parallel
    ! parts of the package.
    ! -----------------------------------------------
    !$ nmaxthreads = omp_get_max_threads()
    !$ call omp_set_num_threads(nmaxthreads)
    ! -----------------------------------------------


    ! Compute latitudes, longitudes and weights of the Gauss--Legendre grid    
    ! associated with the specified "nmax" value
    ! -----------------------------------------------
    print '(a, 1i4, a)', 'Computing the grid nodes of the Gauss--Legendre &
                          &quadrature for degree ', nmax, '...'

    nlat = nmax + 1
    nlon = 2 * nmax + 2

    allocate(lat(nlat), lon(nlon), w(nlat))

    call GL_grid_dp(nmax, &
                    lat, lon, w)
    ! -----------------------------------------------


    ! Import spherical harmonic coefficients from the
    ! "Earth2014.RET2014.degree360.txt" file
    ! -----------------------------------------------
    print *, 'Reading input spherical harmonic coefficients...'


    open(newunit = fid, file = &
         trim(adjustl(path))//'/Earth2014.RET2014.degree360.txt', &
         status = 'old')

    allocate(cs(nmax + 1, nmax + 1))

    do n = 1, (nmax + 1)
        read(fid, *) cs(n, :)
    end do

    close(fid)
    ! -----------------------------------------------


    ! Synthesis of the input coefficients "cs" at the nodes of the
    ! Gauss--Legendre grid
    ! -----------------------------------------------
    print '(a, 1i4, a)', 'Spherical harmonic synthesis from the input &
                          &coefficients up to degree ', nmax, '...'

    allocate(f(nlat, nlon))
    call Surface_SHS_dp(lat, nlat, lon, nlon, nmax, cs, nmax + 1, &
                        f)
    ! -----------------------------------------------






    ! EXAMPLE NO. 1
    ! -----------------------------------------------
    ! Spherical harmonic analysis of the synthesized signal "f"
    ! ...............................................
    print *, ''
    print *, ''
    print *, '========================================'
    print *, ''
    print *, ''
    print *, 'EXAMPLE NO. 1'
    print *, ''
    print '(a, 1i4, a)', 'Spherical harmonic analysis of the synthesized &
                          &signal up to degree ', nmax, '...'


    allocate(cs2(nmax + 1, nmax + 1))

    call Surface_SHA_dp(lat, nlat, nlon, w, f, nmax, nmax, &
                        cs2)
    print *, ''
    ! ...............................................


    ! Validate computed spherical harmonic coefficients with respect to the
    ! input coefficients
    ! ...............................................
    print *, ''
    print *, ''
    print *, 'Validation of the obtained spherical harmonic coefficients:'
    print *, ''
    print *, 'Some sample original coefficients:'
    print *, 'C(50, 20) = ', cs(51, 21)
    print *, 'S(360, 360) = ', cs(360, 361)
    print *, ''
    print *, 'Differences (Delta) between the computed and original sample &
              &coefficients:'
    print *, 'delta C(50, 20) = ', (cs2(51, 21) - cs(51, 21))
    print *, 'delta S(360, 360) = ', (cs2(360, 361) - cs(360, 361))
    print *, ''
    print *, 'max(abs(Delta)) value discrepancy for all coefficients:'
    print *, maxval(abs(cs2 - cs))
    print *, ''
    print *, ''
    ! ...............................................
    ! -----------------------------------------------






    ! EXAMPLE NO. 2
    ! -----------------------------------------------
    ! Now let's use the same input signal "f" that was synthesized up to degree
    ! "nmax = 360", but this time it is harmonically analyzed up to degree
    ! "nmax2 = 50" (instead of 360 as done above). This example demonstrates
    ! that if a band-limited signal is sampled correctly, than spherical 
    ! harmonic coefficients can be computed *exactly* for any degree
    ! "nmax2 = 0,... , nmax" (further details can be found in the
    ! "Surface_SHA.f95" subroutine).
    ! ...............................................
    print *, '========================================'
    print *, ''
    print *, ''
    print *, 'EXAMPLE NO. 2'
    print *, ''
    print '(a, 1i4, a, 1i4, a)', &
          'Spherical harmonic analysis of the synthesized signal, but this &
           &time only up to degree', 50, ', despite that the input signal &
           &possesses harmonics up to degree ', nmax, '...'


    allocate(cs3(51, 51))

    call Surface_SHA_dp(lat, nlat, nlon, w, f, nmax, 50, &
                        cs3)

    deallocate(f)
    ! ...............................................


    ! Validate computed spherical harmonic coefficients with respect to the
    ! input coefficients
    ! ...............................................
    print *, ''
    print *, ''
    print *, 'Validation of the obtained spherical harmonic coefficients:'
    print *, ''
    print *, 'Some sample original coefficients:'
    print *, 'C(30, 20) = ', cs(31, 21)
    print *, 'S(50, 50) = ', cs(50, 51)
    print *, ''
    print *, 'Differences (Delta) between the computed and original sample &
              &coefficients:'
    print *, 'delta C(30, 20) = ', (cs3(31, 21) - cs(31, 21))
    print *, 'delta S(50, 50) = ', (cs3(50, 51) - cs(50, 51))
    print *, ''
    print *, 'max(abs(Delta)) value discrepancy for all coefficients:'
    print *, maxval(abs(cs3 - cs(1:51, 1:51)))
    print *, ''
    print *, ''
    print *, '========================================'
    ! ...............................................
    ! -----------------------------------------------






    ! EXAMPLE NO. 3
    ! -----------------------------------------------
    ! Spherical harmonic synthesis of the mean values of "f" over specified
    ! grid cells
    print *, ''
    print *, ''
    print *, 'EXAMPLE NO. 3'
    print *, ''
    print '(a, 1i4, a)', 'Spherical harmonic synthesis of the mean values &
                          &of the input signal over grid cells &
                          &up to degree ', nmax, '...'

    ! Computational grid
    ! ...............................................
    nlat = nmax + 1       ! Number of latitudes
    nlon = 2 * (nmax + 1) ! Number of longitudes


    allocate(lat1(nlat), lat2(nlat))
    allocate(lon1(nlon), lon2(nlon))


    ! Create vectors "lat1" and "lat2" that specify the minimum and maximum 
    ! latitudes of the grid cells, respectively
    do n = 1, nlat
        lat1(n) = -pi_dp / 2.0_dp + real(n - 1, dp) * pi_dp / real(nlat, dp) 
        lat2(n) = -pi_dp / 2.0_dp + real(n, dp) * pi_dp / real(nlat, dp) 
    end do


    ! Create vectors "lon1" and "lon2" that specify the minimum and maximum 
    ! longitudes of the grid cells, respectively
    do n = 1, nlon
        lon1(n) = real(n - 1, dp) * 2.0_dp * pi_dp / real(nlon, dp) 
        lon2(n) = real(n, dp) * 2.0_dp * pi_dp / real(nlon, dp) 
    end do


    ! Harmonic synthesis of the mean values of "f" over the grid cells
    ! ...............................................
    allocate(f(nlat, nlon))

    call Surface_SHS_mean_dp(lat1, lat2, nlat, lon1, lon2, nlon, nmax, cs, &
                            &nmax + 1, &
                            &f)
    ! ...............................................


    ! Sample outputs from the synthesized mean values
    ! ...............................................
    print *, ''
    print *, ''
    print *, 'Sample outputs for some of the grid cells:'
    print *, ''
    print *, 'lat_min (deg)              lat_max (deg)              &
             &lon_min (deg)             lon_max (deg)            f_mean'
    print *, lat1(15) * 180.0_dp / pi_dp,  lat2(15) * 180.0_dp / pi_dp, &
            &lon1(120) * 180.0_dp / pi_dp, lon2(120) * 180.0_dp / pi_dp, &
            &f(15, 120) 
    print *, lat1(130) * 180.0_dp / pi_dp, lat2(130) * 180.0_dp / pi_dp, &
            &lon1(400) * 180.0_dp / pi_dp, lon2(400) * 180.0_dp / pi_dp, &
            &f(130, 400) 
    print *, lat1(230) * 180.0_dp / pi_dp, lat2(230) * 180.0_dp / pi_dp, &
            &lon1(450) * 180.0_dp / pi_dp, lon2(450) * 180.0_dp / pi_dp, &
            &f(230, 450) 
    print *, ''
    print *, ''
    print *, '========================================'
    ! ...............................................
    ! -----------------------------------------------


    ! EXAMPLE NO. 4
    ! -----------------------------------------------
    ! Spherical harmonic analysis of the mean values of "f" (taken from Example 
    ! No. 3)
    print *, ''
    print *, ''
    print *, 'EXAMPLE NO. 4'
    print *, ''
    print '(a, 1i4, a)', 'Spherical harmonic analysis of the mean values &
                          &of the input signal (taken from Example No. 3) &
                          &up to degree ', nmax, '...'

    allocate(cs4(nmax + 1, nmax + 1))

    call Surface_SHA_mean_dp(lat1, lat2, nlat, lon1, lon2, nlon, f, nmax, &
                            &cs4)
    
    print *, ''
    print *, 'Sample output spherical harmonic coefficients'
    print *, ''

    print '(a, e30.16)', 'C10,5   ', cs4(11, 6)
    print '(a, e30.16)', 'S10,5   ', cs4(5, 11)
    print '(a, e30.16)', 'C100,100', cs4(101, 101)
    print '(a, e30.16)', 'S100,100', cs4(100, 101)
    print '(a, e30.16)', 'C360,360', cs4(361, 361)
    print '(a, e30.16)', 'S360,360', cs4(360, 361)
    print *, ''
    print *, ''
    print *, '========================================'
    ! -----------------------------------------------


    print *, ''
    print *, ''
    print *, 'End of the test computation.'
    ! ===============================================


end program Test_run_dp
