subroutine Surface_SHA_dp(lat, nlat, nlon, w, f, nmax_signal, nmax_sha, cs)
!
! ==========================================================================
!
! DESCRIPTION: This subroutine computes spherical harmonic coefficients of a
!              signal "f" via the exact FFT-based Gauss--Legendre quadrature
!              (Sneeuw, 1994) up to degree "nmax_sha". The signal must be
!              sampled at the nodes of the Gauss--Legendre quadrature (see the
!              "GL_grid.f95" subroutine) and possess harmonics up to degree
!              "nmax_signal >= nmax_sha" (see below).
!
!              The routine is written in double precision.
!
!              Importantly, the subroutine allows to compute spherical harmonic
!              coefficients even for a smaller maximum degree than that of the
!              input signal, that is, for "nmax_sha <= nmax_signal". In that
!              case, however, the input signal must be sampled correspondingly
!              to its maximum harmonic degree "nmax_signal", that is, at the
!              grid obtained using the "nmax_signal" value as the input to the
!              "GL_grid_dp.f95" subroutine. If this condition is met, the
!              resulting coefficients up to degree "nmax_sha" are computed
!              exactly without any approximation errors (neglecting numerical
!              errors, etc.).
!
!              The loop over latitudes is parallelized using OpenMP.
!
!              The 4*pi fully-normalized associated Legendre functions
!              (e.g., Heiskannen and Moritz, 1967) are evaluated by the
!              numerically stable algorithm by Fukushima (2012). The algorithm
!              can be used up to ultra-high harmonic degrees (say, tens of
!              thousands or even well beyond).
!
!              As the main input, the signal to be analyzed is represented
!              by a matrix as one of the input variables (see the variable "f"
!              below). Importantly, the whole matrix is stored in RAM during the
!              entire harmonic analysis. The same holds for the output spherical
!              harmonic coefficients. Ultra-high-degree analysis may therefore
!              require tens of GBs of RAM.
!
!              The subroutine exploits the symmetry property of Legendre
!              functions with respect to the equator, thereby improving its
!              performance, especially for ultra-high-degree harmonic analysis.
!              For further details on the symmetric grid, see
!              "Surface_SHS_dp.f95". However, unlike "Surface_SHS_dp.f95", this
!              subroutine always takes advantage of the symmetry, as the
!              Gauss--Legendre grid is always symmetric and no other grid is
!              currently supported in this subroutine.
!
!
! INPUTS: "lat"         -- Latitudes of the Gauss--Legendre grid in radians.
!                          A vector of the dimension (nmax_signal + 1), where
!                          "nmax_signal" is the maximum degree present in the
!                          signal (see above and the "GL_grid.f95" subroutine).
!
!         "nlat"        -- Number of latitudes in a single meridian
!                          ("nlat = nmax_signal + 1").
!
!         "nlon"        -- Number of longitudes in a single latitude parallel
!                          ("nlon = 2 * nmax_signal + 2").
!
!         "w"           -- Weights due to the Gauss--Legendre quadrature.
!                          A vector of the dimension (nmax_signal + 1). See
!                          above and the "GL_grid_dp.f95" subroutine.
!
!         "f"           -- Input signal i) given at the nodes of the
!                          Gauss--Legendre quadrature for "nmax_signal" and
!                          ii) containing harmonics up to degree
!                          "nmax_signal".
!
!                          A matrix of the dimensions (nmax_signal + 1,
!                          2 * nmax_signal + 2), sampled as follows:
!
!
! [(lat(1),lon(1))              ...               (lat(1),lon(2*nmax_signal+2)]
! [      .                      .                              .              ]
! [      .                       .                             .              ]
! [      .                        .                            .              ]
! [(lat(nmax_signal+1),lon(1))  ...   (lat(nmax_signal+1),lon(2*nmax_signal+2)]
!
!
!                          where "lat" and "lon" are vectors generated by
!                          the "GL_grid_dp.f95" subroutine using the following
!                          command
!
!                          "call GL_grid_dp(nmax_signal, lat, lon, w)"
!
!                          See above and also the "Surface_SHS.f95" subroutine.
!
!         "nmax_signal" -- Maximum harmonic degree present in the signal
!                          (here, the variable "f" above).
!
!         "nmax_sha"    -- Maximum harmonic degree of the output spherical
!                          harmonic coefficients. This value cannot be larger
!                          than "nmax_signal". Note that if
!                          "nmax_sha < nmax_signal", the output coefficients
!                          are still computed exactly, because the matrix
!                          "f" is sampled correctly according to the
!                          maximum degree that is present in the signal, here
!                          "nmax_signal".
!
!
! OUTPUTS: "cs" -- Matrix of dimensions (nmax_sha + 1, nmax_sha + 1) with
!                  output spherical harmonic coefficients. The structure of
!                  of the matrix is as follows:
!
!
!                     [C(0, 0) S(1, 1) S(2, 1)    ...           S(nmax_SHA, 1)]
!                     [C(1, 0) C(1, 1) S(2, 2)    ...           S(nmax_SHA, 2)]
!                     [C(2, 0) C(2, 1) C(2, 2)    ...           S(nmax_SHA, 3)]
!                     [  .       .      .      .                       .      ]
!                     [  .       .      .       .                      .      ]
!                     [  .       .      .        .                     .      ]
!                     [C(nmax_SHA, 0)   ...              C(nmax_SHA, nmax_SHA)]
!
!
!                  where C(n, m) is the coefficient "C" of degree "n" and order
!                  "m", etc.
!
!
! REFERENCES: Sneeuw, N. (1994) Global spherical harmonic analysis by
!                least-squares and numerical quadrature methods in historical
!                perspective. Geophysical Journal International 118:707--716
!
!             Heiskannen, W. A., Moritz, H. (1967) Physical Geodesy. W. H.
!                Freeman and Company, San Francisco, 364 pp
!
!             Fukushima, T. (2012) Numerical computation of spherical
!                harmonics of arbitrary degree and order by extending exponent
!                of floating point numbers. Journal of Geodesy 86:271--285.
!
!
! Contact: blazej.bucha@stuba.sk
!
!
! Please use the following reference when using this subroutine:
!
!          Bucha, B., Hirt, C., Kuhn, M., 2019. Cap integration in
!             spectral gravity forward modelling up to the full gravity
!             tensor. Journal of Geodesy 93:1707--1737
!
! ==========================================================================





    ! Import modules
    ! ===============================================
    use vartypes,  only: dp ! Module defining number formats
    use constants, only: pi_dp, ind_dp, root3_dp ! Module defining constants
                                                 ! (pi, etc.)
    use FFTW3 ! Module for the FFTW package
    ! ===============================================






    ! Declaration of variables
    ! ===============================================
    implicit none

    integer, intent(in)  :: nmax_sha       ! Maximum degree of the output
                                           ! spherical harmonic coefficients
    integer, intent(in)  :: nmax_signal    ! Maximum harmonic degree present in
                                           ! the signal
    integer              :: nlat           ! Number of latitudes in one meridian
    integer              :: nlon           ! Number of longitudes in one
                                           ! latitude parallel
    real(dp), intent(in) :: lat(nlat)      ! Spherical latitudes of the grid
                                           ! points in radians
    real(dp), intent(in) :: w(nlat)        ! Weights for the Gauss--Legendre
                                           ! quadrature
    real(dp), intent(in) :: f(nlat, nlon)  ! Input quantity to be harmonically
                                           ! analyzed


    ! Legendre functions (LFs) and some variables necessary for their evaluation
    ! via the extended-range arithmetic (see Fukushima 2012)
    real(dp), parameter :: big = 2.0_dp**ind_dp
    real(dp), parameter :: bigi = 2.0_dp**(-ind_dp)
    real(dp), parameter :: bigs = 2.0_dp**(ind_dp / 2)
    real(dp), parameter :: bigsi = 2.0_dp**(-ind_dp / 2)
    real(dp) :: pnm0, pnm1, pnm2   ! LFs
    real(dp) :: am(nmax_sha + 1), bm(nmax_sha)
    real(dp) :: dn(nmax_sha), en(nmax_sha + 1), fn(nmax_sha + 1)
    real(dp) :: ps(nmax_sha), r(2 * nmax_sha + 3), ri(2 * nmax_sha + 3)
    real(dp) :: u, t, x, y, z, wlf
    integer  :: ix, iy, iz, ixy, ips(nmax_sha)

    integer :: even    ! If the number of latitudes is even, then "even = 1",
                       ! otherwise "even = 0"
    integer :: nlatdo  ! If the grid is symmetric with respect to the equator,
                       ! then "nlatdo" represents the number of latitudes from
                       ! one of the poles to the equator (including)

    integer :: i, n, m  ! Loop variables over the latitudes, harmonic degrees
                        ! and orders

    real(dp) :: a(nlon / 2 + 1), b(nlon / 2 + 1)   ! Lumped coefficients
    real(dp) :: a2(nlon / 2 + 1), b2(nlon / 2 + 1) ! Lumped coefficients
    real(dp) :: amp1, bmp1    ! Lumped coefficients of order "m"
    real(dp) :: a2mp1, b2mp1  ! Lumped coefficients of order "m"

    real(dp) :: c, cw  ! Some auxiliary variables
    integer  :: imax

    ! Some variables necessary for FFT
    type(c_ptr) :: plan
    real(dp)    :: ftmp_in(nlon)
    complex(dp) :: ftmp_out(nlon / 2 + 1)

    real(dp) :: csatmp, csbtmp  ! Temporary variables for products of LFs and
                                ! lumped coefficients
    real(dp), intent(out) :: cs(nmax_sha + 1, nmax_sha + 1) ! Output matrix of
                                                            ! spherical harmonic
                                                            ! coefficients
    ! ===============================================






    ! The main computational part of the subroutine
    ! ===============================================

    ! Check whether the number of latitudes is even or odd
    ! --------------------------------------------------------------------------
    if (mod(nlat, 2) .eq. 0) then
        even = 1  ! The number of latitudes is an even number. The grid does not
                  ! contain the zero latitude.
        nlatdo = nlat / 2
    else
        even = 0  ! The number of latitudes is an odd number. The grid could
                  ! possibly contain the zero latitude.
        nlatdo = (nlat + 1) / 2
    end if
    
    imax = nlatdo + even
    ! --------------------------------------------------------------------------



    ! Initializations for recurrence relations to compute Legendre functions
    ! --------------------------------------------------------------------------
    do m = 1, (2 * nmax_sha + 3)
        wlf = sqrt(real(m, dp))
        r(m) = wlf
        ri(m) = 1.0_dp / wlf
    end do


    ! Computation of the "dn" coefficients for Legendre recurrence 
    ! relations
    do n = 1, nmax_sha
        dn(n) = r(2 * n + 3) * ri(2 * n + 2)
    end do


    ! Computation of "en" and "fn" coefficients for Legendre recurrence
    ! relations
    do n = 2, nmax_sha
        en(n + 1) = r(2 * n + 1) * r(2 * n - 1) / real(n, dp)
        fn(n + 1) = real(n - 1, dp) * r(2 * n + 1) / (real(n, dp) * &
                   &r(2 * n - 3))
    end do
    ! --------------------------------------------------------------------------


    ! Auxiliary variable entering the computation of the lumped coefficients
    ! --------------------------------------------------------------------------
    c = pi_dp * 2.0_dp / (2.0_dp * (real(nmax_signal, dp) + 1.0_dp))
    ! --------------------------------------------------------------------------


    ! Initialize all elements of the output coefficients to zero
    ! --------------------------------------------------------------------------
    cs = 0.0_dp
    ! --------------------------------------------------------------------------






    ! Create a plan for FFT
    ! --------------------------------------------------------------------------
    plan = fftw_plan_dft_r2c_1d(nlon, ftmp_in, ftmp_out, FFTW_ESTIMATE)
    ! --------------------------------------------------------------------------






    ! Loop over latitudes
    ! --------------------------------------------------------------------------
    !$OMP PARALLEL DO REDUCTION(+:cs) DEFAULT(NONE) &
    !$OMP& SHARED(f, lat, nlat, nlon, w, nlatdo, imax, c, nmax_sha) &
    !$OMP& SHARED(dn, en, fn, r, ri, plan) &
    !$OMP& PRIVATE(pnm0, pnm1, pnm2, t, u) &
    !$OMP& PRIVATE(x, ix, y, iy, z, iz, ps, ips, ixy) &
    !$OMP& PRIVATE(a, b, a2, b2, am, bm, amp1, a2mp1, bmp1, b2mp1) &
    !$OMP& PRIVATE(csatmp, csbtmp, ftmp_in, ftmp_out) &
    !$OMP& PRIVATE(cw, wlf)
    do i = 1, nlatdo

        ! Lumped coefficients for the southern hemisphere (including the
        ! equator)
        ! ------------------------------------------------------------------
        ftmp_in = f(i, :)
        call fftw_execute_dft_r2c(plan, ftmp_in, ftmp_out)

        cw =   c * w(i)
        a  =  cw * real(ftmp_out, dp)
        b  = -cw * aimag(ftmp_out)
        ! ------------------------------------------------------------------


        ! Lumped coefficients for the northern hemisphere
        ! ------------------------------------------------------------------
        if (i < imax) then
            ftmp_in = f(nlat - i + 1, :)
            call fftw_execute_dft_r2c(plan, ftmp_in, ftmp_out)

            cw =   c * w(nlat - i + 1)
            a2 =  cw * real(ftmp_out, dp)
            b2 = -cw * aimag(ftmp_out)
        end if
        ! ------------------------------------------------------------------


        ! Some precomputations for Legendre functions
        ! ------------------------------------------------------------------
        t  = sin(lat(i))
        u  = cos(lat(i))
        x  = root3_dp * u
        ix = 0
         ps(1) =  x
        ips(1) = ix

        do n = 2, nmax_sha
            x = (dn(n - 1) * u) * x
            y = abs(x)

            if (y >= bigs) then
                 x =  x * BIGI
                ix = ix + 1
            elseif (y < bigsi) then
                 x =  x * big
                ix = ix - 1
            end if

             ps(n) =  x
            ips(n) = ix
        end do
        ! ------------------------------------------------------------------






        ! Loop over harmonic orders
        ! ------------------------------------------------------------------
        do m = 0, nmax_sha


            ! Computation of "am" and "bm" coefficients for Legendre recurrence
            ! relations
            ! ------------------------------------------------------------------
            am(m + 1) = r(2 * m + 3)
            do n = (m + 2), nmax_sha
                wlf   = r(2 * n + 1) * ri(n - m) * ri(n + m)
                am(n) = r(2 * n - 1) * wlf
                bm(n) = r(n - m - 1) * r(n + m - 1) * ri(2 * n - 3) * wlf
            end do
            ! ------------------------------------------------------------------


            ! Some useful substitutions
            ! ------------------------------------------------------------------
            amp1  = a(m + 1)
            a2mp1 = a2(m + 1)
            bmp1  = b(m + 1)
            b2mp1 = b2(m + 1)
            ! ------------------------------------------------------------------


            ! Computation of spherical harmonic coefficients
            ! ------------------------------------------------------------------
            if (m == 0) then

                ! Zonal harmonics
                ! --------------------------------------------------------------
                ! P00
                pnm0 = 1.0_dp
                csatmp = pnm0 * amp1

                if (i < imax) then
                    csatmp = csatmp + pnm0 * a2mp1
                end if
                cs(1, m + 1) = cs(1, m + 1) + csatmp ! C00

                ! P10
                if (nmax_sha >= 1) then

                    pnm1   = root3_dp * t
                    csatmp = pnm1 * amp1

                    if (i < imax) then
                        csatmp = csatmp - pnm1 * a2mp1
                    end if
                    cs(2, m + 1) = cs(2, m + 1) + csatmp ! C10

                end if

                ! P20, P30, ..., Pnmax,0
                if (nmax_sha >= 2) then

                    do n = 2, nmax_sha

                        pnm2   = en(n + 1) * t * pnm1 - fn(n + 1) * pnm0
                        csatmp = pnm2 * amp1

                        if (i < imax) then
                            if (mod(n, 2) == 0) then
                                csatmp = csatmp + pnm2 * a2mp1
                            else
                                csatmp = csatmp - pnm2 * a2mp1
                            end if
                        end if
                        cs(n + 1, m + 1) = cs(n + 1, m + 1) + csatmp ! C20,
                                                                     ! C30, ...,
                                                                     ! Cnmax,0

                        pnm0 = pnm1
                        pnm1 = pnm2

                    end do

                end if
                ! --------------------------------------------------------------

            else ! Non-zonal harmonics

                ! Sectorial harmonics
                ! --------------------------------------------------------------
                x = ps(m)
                ix = ips(m)

                ! Pmm
                if (ix == 0) then
                    pnm0 = x
                elseif (ix < -1) then
                    pnm0 = 0.0_dp
                elseif (ix < 0) then
                    pnm0 = x * bigi
                else
                    pnm0 = x * big
                end if

                csatmp = pnm0 * amp1
                csbtmp = pnm0 * bmp1

                if (i < imax) then
                    csatmp = csatmp + pnm0 * a2mp1
                    csbtmp = csbtmp + pnm0 * b2mp1
                end if

                cs(m + 1, m + 1) = cs(m + 1, m + 1) + csatmp ! Cmm
                cs(m, m + 1)     = cs(m, m + 1)     + csbtmp ! Smm
                ! --------------------------------------------------------------


                ! Tesseral harmonics
                ! --------------------------------------------------------------
                if (m < nmax_sha) then
                     y =  x
                    iy = ix
                     x = (am(m + 1) * t) * y
                    ix = iy

                    wlf = abs(x)
                    if (wlf >= bigs) then
                         x =  x * bigi
                        ix = ix + 1
                    elseif (wlf < bigsi) then
                         x =  x * big
                        ix = ix - 1
                    end if

                    ! Pm+1,m
                    if (ix == 0) then
                        pnm1 = x
                    elseif (ix < - 1) then
                        pnm1 = 0.0_dp
                    elseif (ix < 0) then
                        pnm1 = x * bigi
                    else
                        pnm1 = x * big
                    end if

                    csatmp = pnm1 * amp1
                    csbtmp = pnm1 * bmp1

                    if (i < imax) then
                        csatmp = csatmp - pnm1 * a2mp1
                        csbtmp = csbtmp - pnm1 * b2mp1
                    end if

                    cs(m + 2, m + 1) = cs(m + 2, m + 1) + csatmp ! Cm+1,m
                    cs(m, m + 2)     = cs(m, m + 2)     + csbtmp ! Sm+1,m

                    ! Loop over degrees
                    ! ----------------------------------------------------------
                    do n = (m + 2), nmax_sha

                        ixy = ix - iy

                        if (ixy == 0) then
                             z = (am(n) * t) * x - bm(n) * y
                            iz = ix
                        elseif (ixy == 1) then
                             z = (am(n) * t) * x - bm(n) * (y * bigi)
                            iz = ix
                        elseif (ixy == -1) then
                             z = (am(n) * t) * (x * bigi) - bm(n) * y
                            iz = iy
                        elseif (ixy > 1) then
                             z = (am(n) * t) * x
                            iz = ix
                        else
                             z = -bm(n) * y
                            iz = iy
                        end if

                        wlf = abs(z)
                        if (wlf >= bigs) then
                             z =  z * bigi
                            iz = iz + 1
                        elseif (wlf < bigsi) then
                             z =  z * big
                            iz = iz - 1
                        end if

                        ! Pm+2,m, Pm+3,m, ..., Pnmax,m
                        if (iz == 0) then
                            pnm2 = z
                        elseif (iz < -1) then
                            pnm2 = 0.0_dp
                        elseif (iz < 0 ) then
                            pnm2 = z * bigi
                        else
                            pnm2 = z * big
                        end if

                        csatmp = pnm2 * amp1
                        csbtmp = pnm2 * bmp1

                        if (i < imax) then

                            if (mod(n + m, 2) == 0) then
                                csatmp = csatmp + pnm2 * a2mp1
                                csbtmp = csbtmp + pnm2 * b2mp1
                            else
                                csatmp = csatmp - pnm2 * a2mp1
                                csbtmp = csbtmp - pnm2 * b2mp1
                            end if

                        end if

                        cs(n + 1, m + 1) = cs(n + 1, m + 1) + csatmp ! Cm+2,m,
                                                                     ! Cm+3,m,
                                                                     ! ...,
                                                                     ! Cnmax,m
                        cs(m, n + 1)     = cs(m, n + 1)     + csbtmp ! Sm+2,m,
                                                                     ! Sm+3,m,
                                                                     ! ...,
                                                                     ! Snmax,m

                         y =  x
                        iy = ix
                         x =  z
                        ix = iz

                    end do ! End of the loop over harmonic degrees
                    ! ----------------------------------------------------------

                end if ! End of computation of tesseral harmonics
                ! --------------------------------------------------------------

            end if ! End of computation of spherical harmonic coefficients
            ! ------------------------------------------------------------------

        end do ! End of the loop over harmonic orders
        ! ------------------------------------------------------------------

    end do ! End of the loop over grid latitudes
    !$OMP END PARALLEL DO

    call fftw_destroy_plan(plan)
    ! ------------------------------------------------------------------

    cs = cs / (4.0_dp * pi_dp)
    ! ===============================================

end subroutine Surface_SHA_dp
