subroutine Surface_SHS_mean_qp(lat1, lat2, nlat, lon1, lon2, nlon, nmax, cs, &
                              &ncs, f)
!
! ==========================================================================
!
! DESCRIPTION: This subroutine computes mean values of some function that is
!              expanded into surface spherical harmonics by computing its
!              integral on the unit sphere over grid cells specified by minimum 
!              and maximum latitudes and longitudes. The subroutine employs a 
!              FFT-based algorithm (e.g., Colombo, 1981).
!
!              The routine is written in quadruple precision.
!
!              The loop over latitude cells is parallelized using OpenMP.
!
!              The subroutine automatically starts the synthesis from degree
!              "nmin = 0". If you wish to use some other value of "nmin"
!              (but satisfying "1 <= nmin <= nmax"), you can set (outside this
!              subroutine) the respective coefficients of the "cs" matrix to
!              zero via
!
!              cs(1:nmin, 1:nmin) = 0.0_qp
!
!              where "nmin" is an integer satisfying "1 <= nmin <= nmax", and
!              then use the new matrix as the input to this subroutine. In that
!              case, the synthesis still starts at degree 0, but since the
!              coefficients up to degree "nmin - 1" are zero, they do
!              not contribute to the output signal as required.
!
!              The 4*pi fully-normalized associated Legendre functions
!              (e.g., Heiskannen and Moritz, 1967) are evaluated by the
!              numerically stable algorithm by Fukushima (2012). The algorithm
!              can be used up to ultra-high harmonic degrees (say, tens of
!              thousands or even well beyond).
!
!              The subroutine exploits the symmetry property of Legendre
!              functions and their intergrals with respect to the equator, 
!              thereby improving its performance, especially for        
!              ultra-high-degree harmonic synthesis. This computational 
!              acceleration is employed automatically, provided that the grid 
!              is recognized as symmetric with respect to the equator. Below, 
!              we provide some examples of symmetric and non-symmetric 
!              grids. Note that while these examples use the unit degrees for a 
!              more intuitive understanding, the subroutine requires radians as 
!              the input unit.
!
!              Examples of some symmetric grids (shown are only the minimum and 
!              the maximum latitudes "lat1" and "lat2", respectively):
!              lat1 = [-90, -60, -30,  0, 30, 60]
!              lat2 = [-60, -30,   0, 30, 60, 90]
!
!              lat1 = [-35, -25, -15, -5,  5, 15, 25]
!              lat2 = [-25, -15,  -5,  5, 15, 25, 35]
!
!              lat1 = [-90, -85, 80, 85]
!              lat2 = [-85, -80, 85, 90]
!
!              lat1 = [-90, -80, -75, -70, 69, 70, 75, 80] ! Varying cell size
!              lat2 = [-80, -75, -70, -69, 70, 75, 80, 90] ! Varying cell size
!
!              Examples of some grids that are not considered as symmetric
!              (shown are again only latitudes):
!              lat1 = [-90, -60, -30,  0, 30]
!              lat2 = [-60, -30,   0, 30, 60] ! The negative latitude of -90
!                                             ! deg does not have its positive
!                                             ! counterpart
!
!              lat1 = [-35, -25, -15,     5, 15, 25]
!              lat2 = [-25, -15,  -4.99, 15, 25, 35] ! The difference
!                                                    ! "abs(abs(-4.99)-5)" is
!                                                    ! larger than the
!                                                    ! threshold "thold"
!                                                    ! (see below)
!
!              In other words, the grid is symmetric, provided that all
!              positive latitudes have their negative counterpart or vice versa
!              (up to a given threshold to suppress numerical inaccuracies, see
!              the "thold" variable).
!
!              The longitudes of the grid must be sampled with equal spacing,
!              that is, the difference between any two consecutive longitudes
!              must be constant. The subroutine performs a check on this.
!              If the longitudes does not pass the test, a message is printed
!              and the execution is stopped. If you know for sure that
!              your longitudes share an equal spacing, but never pass
!              the test, you may want to use a larger value of the "thold"
!              variable (see below). This might happen due to the rounding
!              errors with a long vector of longitudes.
!
!
!
! INPUTS: "lat1"      -- Column vector defining the minumum latitudes of the
!                        grid cells in radians (latitudes over a single
!                        meridian).
!
!         "lat2"      -- Column vector defining the maximum latitudes of the
!                        grid cells in radians (latitudes over a single
!                        meridian).
!
!         "nlat"      -- Number of latitudes in the "lat1" and "lat2" arrays.
!
!         "lon1"      -- Column vector defining the minimum longitudes of the
!                        grid cells in radians (longitudes over a single
!                        latitude parallel).
!
!         "lon2"      -- Column vector defining the maximum longitudes of the
!                        grid cells in radians (longitudes over a single
!                        latitude parallel).
!
!         "nlon"      -- Number of longitudes in the "lon1" and "lon2" arrays.
!
!         "nmax"      -- Maximum spherical harmonic degree of the synthesis
!                        (Non-negative integer, 0 <= nmax < ncs). See below for
!                        the explanation of the "ncs" variable.
!
!         "cs"        -- Matrix of dimensions (ncs, ncs) with
!                        input spherical harmonic coefficients (see below
!                        the explanation of "ncs"). The structure
!                        of the matrix is as follows:
!
!                        [C(0, 0) S(1, 1) S(2, 1)    ...          S(ncs - 1, 1)]
!                        [C(1, 0) C(1, 1) S(2, 2)    ...          S(ncs - 1, 2)]
!                        [C(2, 0) C(2, 1) C(2, 2)    ...          S(ncs - 1, 3)]
!                        [  .       .      .         .                  .      ]
!                        [  .       .      .          .                 .      ]
!                        [  .       .      .           .                .      ]
!                        [C(ncs - 1, 0)              ...    C(ncs - 1, ncs - 1)]
!
!                        where C(n, m) is the coefficient "C" of degree "n" and
!                        order "m", etc.
!
!
! OUTPUTS: "f" -- Matrix of the mean values of the synthesized signal (e.g.,
!                 topographic heights) over the specified cells of dimensions
!                 (nlat, nlon) and with the structure
!
!            [(lat(1,1),lon(1,1))         ...         (lat(1,1),lon(nlon,nlon)]
!            [      .                     .                    .              ]
!            [      .                      .                   .              ]
!            [      .                       .                  .              ]
!            [(lat(nlat,nlat),lon(1,1))   ...   (lat(nlat,nlat),lon(nlon,nlon)]
!
!                 where "lat(i,i)" represents the cell with the minimum 
!                 latitude "lat1(i)" and maximum latitude "lat2(i)", and 
!                 "lon(j,j)" stands for the cell with the minimum longitude 
!                 "lon1(j)" and maximum longitude "lon2(j)".
!
!
! REFERENCES: Colombo, O. L. (1981) Numerical methods for harmonic analysis on 
!                the sphere. Reports of the Department of Geodetic Science, 
!                Report No. 310, 140 pp
!
!             Heiskannen, W. A., Moritz, H. (1967) Physical Geodesy. W. H.
!                Freeman and Company, San Francisco, 364 pp
!            
!             Jekeli, C., Lee, J. K., Kwon, J. H. (2007) On the computation and 
!               approximation of ultra-high-degree spherical harmonic 
!               series. Journal of Geodesy 81:603--615
!
!             Fukushima, T. (2012) Numerical computation of spherical
!                harmonics of arbitrary degree and order by extending exponent
!                of floating point numbers. Journal of Geodesy 86:271--285.
!
!             Balmino, Vales, Bonvalot, Briais, 2012. Spherical harmonic
!                 modelling to ultra-high degree of Bouguer and isostatic
!                 anomalies. Journal of Geodesy 86:499–520,
!                 doi: 10.1007/s00190-011-0533-4
!
!             Fukushima, T. (2014) Numerical computation of spherical harmonics
!                 of arbitrary degree and order by extending exponent of
!                 floating point numbers: III integral. Computers & Geosciences
!                 63:17--21
!
!
! Contact: blazej.bucha@stuba.sk
!
!
! Please use the following reference when using this subroutine:
!
!          Bucha, B., Hirt, C., Kuhn, M., 2019. Cap integration in
!             spectral gravity forward modelling up to the full gravity
!             tensor. Journal of Geodesy 93:1707--1737
!
! =========================================================================






    ! Import modules
    ! ===============================================
    use vartypes,  only: qp ! Module defining number formats
    use constants, only: pi_qp, ind_qp, root3_qp, eps_qp ! Module defining
                                                         ! constants  (pi, 
                                                         ! etc.)                                                
    use FFTW3 ! Module for the FFTW package
    ! ===============================================






    ! Declaration of variables
    ! ===============================================
    implicit none

    integer,  intent(in) :: nmax ! Maximum degree of SH expansion
    integer,  intent(in) :: nlat ! Number of grid cells in the latitudinal
                                 ! direction
    integer,  intent(in) :: nlon ! Number of grid cells in the longitudinal
                                 ! direction
    real(qp), intent(in) :: lat1(nlat), lon1(nlon)  ! Minimum spherical
                                                    ! latitudes and longitudes ! of the grid cells in 
                                                    ! radians
    real(qp), intent(in) :: lat2(nlat), lon2(nlon)  ! Maximum spherical
                                                    ! latitudes and longitudes 
                                                    ! of the grid cells in 
                                                    ! radians
    integer,  intent(in) :: ncs  ! Number of rows of the "cs" matrix (equals to
                                 ! number of columns of "cs")
    real(qp), intent(in) :: cs(ncs, ncs)  ! Matrix of spherical
                                          ! harmonic coefficients


    ! Legendre functions (LFs) and their integrals (I) and some variables 
    ! necessary for their evaluation via the extended-range arithmetic approach
    real(qp), parameter :: big = 2.0_qp**ind_qp
    real(qp), parameter :: bigi = 2.0_qp**(-ind_qp)
    real(qp), parameter :: bigs = 2.0_qp**(ind_qp / 2)
    real(qp), parameter :: bigsi = 2.0_qp**(-ind_qp / 2)
    real(qp) :: lat1i, lat2i
    real(qp) :: pnm0_lat1i, pnm0_lat2i   ! LFs
    real(qp) :: pnm1_lat1i, pnm1_lat2i   ! LFs
    real(qp) :: pnm2_lat1i, pnm2_lat2i   ! LFs
    real(qp) :: lon(nlon) ! "lon1 + lon2"
    real(qp) :: am(nmax + 1), bm(nmax + 1), dn(nmax + 1)
    real(qp) :: en((nmax + 1) + 1), fn((nmax + 1) + 1)
    real(qp) :: gn(nmax + 1), hn(nmax + 1)
    real(qp) :: ps1((nmax + 1)), ps2((nmax + 1))
    real(qp) :: r(2 * (nmax + 1) + 3), ri(2 * (nmax + 1) + 3)
    real(qp) :: u1, t1, x1, y1, z1, u2, t2, x2, y2, z2, w
    integer  :: ix1, ix2, iy1, iy2, iz1, iz2, ixy1, ixy2
    integer  :: ips1((nmax + 1)), ips2((nmax + 1))
    real(qp) :: in0, imm0, imm1, imm2, inm0, inm1, inm2 ! Integrals of Legendre
                                                        ! functions

    ! Some variables for the PSLR algorithm (see below)
    real(qp) :: m2, m2sm2dl
    real(qp) :: dm0, dm1, dm2
    real(qp) :: dm02, dm12, dm22
    real(qp) :: lon0, dlon, cmdlon2
    real(qp) :: lontmp, clontmp, slontmp

    real(qp) :: dsigma ! The cell area on the unit sphere

    real(qp) :: dlonchck ! Auxiliary variable to check the longitude step
                         ! of the grid

    integer :: even    ! If the number of latitudes is even, then "even = 1",
                       ! otherwise "even = 0"
    integer :: symm    ! If the grid is symmetric with respect to the equator,
                       ! then "symm = 1", otherwise "symm = 0". If "symm == 1",
                       ! the subroutine automatically exploits the symmetry
                       ! property of Legendre functions in order to accelerate
                       ! the computation
    integer :: nlatdo  ! If the grid is symmetric with respect to the equator,
                       ! then "nlatdo" represents the number of latitudes from
                       ! one of the poles to the equator (including)
    logical :: symmi   ! Just some auxiliary variable

    integer :: i, j, n, m  ! Loop variables over the latitudes, longitudes,
                           ! harmonic degrees and orders

    real(qp), parameter :: thold = 10000.0_qp * eps_qp
        ! If abs(a - b) <= thold, then "a" is considered in this subroutine to
        ! be equal to "b" within the threshold "thold". "thold" is used to check
        ! whether input the latitudes and longitudes meet some criteria
        ! (symmetric latitudes with respect to the equator, etc.).
        ! The value of "thold" can be modified by the user if necessary.
        ! The larger the value, the larger numerical inaccuracies will be
        ! tolerated.

    real(qp) :: a, b, a2, b2        ! Lumped coefficients
    real(qp) :: inmcsa, inmcsb      ! Legendre functions times spherical
                                    ! harmonic coefficients to compute lumped
                                    ! coefficients
    real(qp) :: fi(nlon), fi2(nlon) ! Temporary contribution of the ith
                                    ! latitude parallel to the output quantity

    real(qp), intent(out) :: f(nlat, nlon) ! Output synthesized quantity
    ! ===============================================







    ! The main computational part of the subroutine
    ! ===============================================

    ! Check the latitudes of the computational grid
    ! --------------------------------------------------------------------------
    ! Check whether the number of latitudes in "lat1" and "lat2" is even or odd
    ! ..........................................................................
    if (mod(nlat, 2) .eq. 0) then
        even = 1  ! The number of cells in the latitudinal direction is an
                  ! even number
    else
        even = 0  ! The number of cells in the latitudinal direction is an
                  ! odd number
    end if
    ! ..........................................................................


    ! Determine whether the cells in the latitudinal direction are symmetric
    ! with respect to the equator
    ! ..........................................................................
    if (nlat == 1) then ! If there is only one cell in the latitudinal
                        ! direction within the grid, the grid is automatically
                        ! considered as not symmetric
        symm = 0
    else
        symm = 1  ! If there is more than one cell in the latidudinal direction
                  ! in the grid, let's start by assuming that the grid is
                  ! symmetric with respect to the equator and check whether
                  ! this is indeed true
    end if


    do i = 1, nlat
        if (abs(lat1(i) + lat2(nlat - i + 1)) > thold) then

            symm = 0  ! The grid is not symmetric.

            exit  ! Exiting the loop with the final decision: the grid is not
                  ! symmetric with respect to the equator ("symm = 0")
        end if
    end do
    ! ..........................................................................


    ! Finally, if the grid is symmetric, we modify the number of latitudes
    ! "nlat" to be equal to the number of latitudes on one hemisphere only
    ! (including the equatorial cells if present). This is because the
    ! time-consuming "do loop" over evaluation points now needs to run for one
    ! hemisphere only, while the results for the other hemishere are obtained
    ! by exploiting the symmetry property of Legendre functions. This reduces
    ! the number of Legendre functions that need to be evaluated by a factor of
    ! ~2, so saves some computational time
    ! ..........................................................................
    if (symm == 1) then
        if (even == 1) then
            nlatdo = nlat / 2
        else
            nlatdo = (nlat + 1) / 2
        end if
    else
        nlatdo = nlat
    end if
    ! ..........................................................................
    ! --------------------------------------------------------------------------






    ! Check the longitudes of the computational grid
    ! --------------------------------------------------------------------------
    ! Check whether the first elements of "lon1" and "lon2" are the smallest
    ! one. Otherwise, the execution of the code is terminated.
    if (abs(lon1(1) - minval(lon1)) > thold) then
        error stop '[Surface_SHS_mean_qp.f95 says:] The first element of the &
                  &longitude vector "lon1" has to be the smallest one. &
                  &Terminating the code.'
    end if
    if (abs(lon2(1) - minval(lon2)) > thold) then
        error stop '[Surface_SHS_mean_qp.f95 says:] The first element of the &
                  &longitude vector "lon2" has to be the smallest one. &
                  &Terminating the code.'
    end if


    ! Useful substitution
    lon = (lon1 + lon2) / 2.0_qp


    ! Get the origin of the longitude "lon" vector (will be necessary later for
    ! the PSLR algorithm)
    lon0 = lon(1) 


    ! Get the lon step of the grid (will be necessary later for the PSLR
    ! algorithm)
    if (nlon > 1) then
        if (abs((lon1(2) - lon1(1)) - (lon2(2) - lon2(1))) > thold) then
            error stop '[Surface_SHS_mean_qp.f95 says:] The difference &
                      &"lon1(2) - lon1(1)" has to be equal to &
                      &"lon2(2) - lon2(1)".'
        end if
        dlon = lon1(2) - lon1(1)
    else
        dlon = 0.0_qp
    end if


    ! Check whether the longitude step is constant for the entire "lon1" and
    ! "lon2" vectors. Otherwise, the execution of the code is terminated.
    do j = 2, nlon
        dlonchck = lon1(j) - lon1(j - 1)
        if (abs(dlon - dlonchck) > thold) then
            error stop '[Surface_SHS_mean_qp.f95 says:] The step of "lon1" &
                  &has to be constant. Terminating the code.'
        end if

        dlonchck = lon2(j) - lon2(j - 1)
        if (abs(dlon - dlonchck) > thold) then
            error stop '[Surface_SHS_mean_qp.f95 says:] The step of "lon2" &
                  &has to be constant. Terminating the code.'
        end if
    end do
    ! --------------------------------------------------------------------------






    ! Initializations for recurrence relations to compute Legendre functions
    ! --------------------------------------------------------------------------
    do m = 1, (2 * (nmax + 1) + 3)
        w = sqrt(real(m, qp))
        r(m) = w
        ri(m) = 1.0_qp / w
    end do


    ! Coefficients "dn" for recurrence relations to compute fully-normalized 
    ! Legendre functions
    do n = 1, (nmax + 1)
        dn(n) = r(2 * n + 3) * ri(2 * n + 2)
    end do


    ! Coefficients "en" and "fn" for recurrence relations to compute 
    ! *un-normalized* Legendre polynomials. (Note that to compute integrals of 
    ! fully-normalized Legendre polynomials, we need un-normalized Legendre 
    ! polynomials; see, e.g., Eq. B.24 of Jekeli et al. 2007)
    do n = 2, (nmax + 1)
        en(n + 1) = real(2 * n - 1, qp) / real(n, qp)
        fn(n + 1) = real(n - 1, qp) / real(n, qp)
    end do
    ! --------------------------------------------------------------------------








    ! Initializations for recurrence relations to compute integrals of 
    ! fully-normalized Legendre functions
    ! --------------------------------------------------------------------------
    ! Computation of "gn" coefficients
    do n = 3, nmax
        gn(n + 1) = 1.0_qp / real(2 * n + 2, qp) * r(n) * r(2 * n + 1) * &
                    r(2 * n - 1) * ri(n - 1)
    end do

    ! Computation of "hn" coefficients
    do n = 0, nmax
        hn(n + 1) = real(n - 2, qp) / real(n + 1, qp)
    end do
    ! -------------------------------------------------------------------------- 







    ! Initialize all elements of the output signal "f" to zero
    ! --------------------------------------------------------------------------
    f = 0.0_qp
    ! --------------------------------------------------------------------------







    ! Loop over the latitudinal direction of the grid cells
    ! --------------------------------------------------------------------------
    !$OMP PARALLEL DO REDUCTION(+:f) DEFAULT(NONE) &
    !$OMP& SHARED(cs, nmax, lat1, lat2, dn, en, fn, r, ri, nlat, nlon) &
    !$OMP& SHARED(nlatdo, gn, hn)&
    !$OMP& SHARED(lon0, dlon, even, symm) &
    !$OMP& PRIVATE(lat1i, pnm0_lat1i, pnm1_lat1i, pnm2_lat1i) &
    !$OMP& PRIVATE(lat2i, pnm0_lat2i, pnm1_lat2i, pnm2_lat2i) &
    !$OMP& PRIVATE(in0, imm0, imm1, imm2, inm0, inm1, inm2) &
    !$OMP& PRIVATE(t1, u1, x1, ix1, y1, iy1, ixy1, z1, iz1) &
    !$OMP& PRIVATE(t2, u2, x2, ix2, y2, iy2, ixy2, z2, iz2) &
    !$OMP& PRIVATE(ps1, ips1, ps2, ips2, fi, fi2, am, w, bm) &
    !$OMP& PRIVATE(a, b, a2, b2, inmcsa, inmcsb, lontmp) &
    !$OMP& PRIVATE(clontmp, slontmp, dm0, dm1, dm12, dm2, dm02, cmdlon2) &
    !$OMP& PRIVATE(dm22, m2sm2dl, m2, dsigma, symmi)
    do i = 1, nlatdo

        ! Check whether the symmetry property of LFs needs to be exploited
        ! ------------------------------------------------------------------
        if (nlat == 1) then ! There is only a single latitude parallel
                            ! in the grid, so no symmetry property
            symmi = .false.
        elseif (symm == 0) then ! The grid was already classified as
                                ! non-symmetric
            symmi = .false.
        elseif (symm == 1 .and. even == 0 .and. i == nlatdo) then
            symmi = .false.  ! For a symmetric grid containing the equator,
                             ! this ensures that the points on the equator
                             ! will not be processed twice
        else
            symmi = .true. ! Exploit the symmetry property
        end if
        ! ------------------------------------------------------------------


        ! Some precomputations for Legendre functions
        ! ------------------------------------------------------------------
        ! lat1(i)
        ! ..................................................................
        lat1i = lat1(i)
        t1  = sin(lat1i)
        u1  = cos(lat1i)
        x1  = root3_qp * u1
        ix1 = 0

         ps1(1) =  x1
        ips1(1) = ix1

        do n = 2, (nmax + 1)
            x1 = (dn(n - 1) * u1) * x1
            y1 = abs(x1)

            if (y1 >= bigs) then
                 x1 =  x1 * bigi
                ix1 = ix1 + 1
            elseif (y1 < bigsi) then
                 x1 =  x1 * big
                ix1 = ix1 - 1
            end if

             ps1(n) =  x1
            ips1(n) = ix1
        end do
        ! ..................................................................


        ! lat2(i)
        ! ..................................................................
        lat2i = lat2(i)
        t2  = sin(lat2i)
        u2  = cos(lat2i)
        x2  = root3_qp * u2
        ix2 = 0

         ps2(1) =  x2
        ips2(1) = ix2

        do n = 2, (nmax + 1)
            x2 = (dn(n - 1) * u2) * x2
            y2 = abs(x2)

            if (y2 >= bigs) then
                 x2 =  x2 * bigi
                ix2 = ix2 + 1
            elseif (y2 < bigsi) then
                 x2 =  x2 * big
                ix2 = ix2 - 1
            end if

             ps2(n) =  x2
            ips2(n) = ix2
        end do
        ! ..................................................................
        ! ------------------------------------------------------------------


        ! The "fi" vector represents the synthesized quantity "f" for the ith
        ! latitude parallel. Therefore, it needs to be reinitialized to zero
        ! for each "ith" latitude. The same holds true for "fi2".
        fi  = 0.0_qp
        fi2 = 0.0_qp


        ! Loop over harmonic orders
        ! ----------------------------------------------------------------------
        do m = 0, nmax

            ! Computation of "am" and "bm" coefficients for Legendre recurrence
            ! relations
            ! ------------------------------------------------------------------
            am(m + 1) = r(2 * m + 3)
            do n = (m + 2), (nmax + 1)
                w     = r(2 * n + 1) * ri(n - m) * ri(n + m)
                am(n) = r(2 * n - 1) * w
                bm(n) = r(n - m - 1) * r(n + m - 1) * ri(2 * n - 3) * w
            end do
            ! ------------------------------------------------------------------


            if (m == 0) then
                    





                ! Zonal Legendre functions and their integrals
                ! -------------------------------------------------------------
                ! P00
                ! .............................................................
                pnm0_lat1i = 1.0_qp
                pnm0_lat2i = 1.0_qp
                ! .............................................................


                ! P10
                ! .............................................................
                pnm1_lat1i = t1
                pnm1_lat2i = t2
                ! .............................................................


                ! I00
                ! .............................................................
                in0 = t2 - t1
                ! .............................................................
                

                ! Lumped coefficients
                ! .............................................................
                a = in0 * cs(1, m + 1)
                b = 0.0_qp

                if (symmi .eqv. .true.) then
                    a2 = a
                    b2 = 0.0_qp
                end if
                ! .............................................................


                if ((nmax + 1) >= 2) then
    
                    do n = 1, nmax 
                        ! P20, P30, ..., Pnmax+1,0
                        ! Since we use locally the "n" variable to compute 
                        ! Legendre polynomials of degree "n + 1", the "n + 2"th 
                        ! elements have to be taken from the vectors "en" and 
                        ! "fn" below. Once again, note that when "m == 0", then 
                        ! "pnm0", "pnm1" and "pnm2" for lat1i and lat2i 
                        ! represent un-normalized Legendre polynomials. These 
                        ! are needed to get the integrals of fully-normalized 
                        ! Legendre polynomials
                        ! .....................................................
                        pnm2_lat1i = en(n + 2) * t1 * pnm1_lat1i - &
                                     fn(n + 2) * pnm0_lat1i
                        pnm2_lat2i = en(n + 2) * t2 * pnm1_lat2i - &
                                     fn(n + 2) * pnm0_lat2i
                        ! .....................................................
    

                        ! I10, I20, ..., Inmax,0
                        ! Computed from Pn+1,0 and Pn-1,0
                        ! .....................................................
                        in0 = ri(2 * n + 1) * (pnm2_lat2i - pnm0_lat2i - &
                                               pnm2_lat1i + pnm0_lat1i)
                        ! .....................................................


                        ! Lumped coefficients
                        ! .....................................................
                        inmcsa = in0 * cs(n + 1, m + 1)
                        a      = a + inmcsa

                        if (symmi .eqv. .true.) then
                            if (mod(n, 2) == 0) then
                                a2 = a2 + inmcsa
                            else
                                a2 = a2 - inmcsa
                            end if
                        end if
                        ! .....................................................


                        pnm0_lat1i = pnm1_lat1i
                        pnm1_lat1i = pnm2_lat1i


                        pnm0_lat2i = pnm1_lat2i
                        pnm1_lat2i = pnm2_lat2i

                    end do
    
                end if
                ! -------------------------------------------------------------
                





            else ! Non-zonal harmonics





                 
                ! Sectorial Legendre functions and their integrals
                ! -------------------------------------------------------------

                ! Pmm for lat1i
                ! .............................................................
                x1  =  ps1(m)
                ix1 = ips1(m)
    
                if (ix1 == 0) then
                    pnm0_lat1i = x1
                elseif (ix1 < -1) then
                    pnm0_lat1i = 0.0_qp
                elseif (ix1 < 0) then
                    pnm0_lat1i = x1 * bigi
                else
                    pnm0_lat1i = x1 * big
                end if
                ! .............................................................


                ! Pmm for lat2i
                ! .............................................................
                x2  =  ps2(m)
                ix2 = ips2(m)
    
                if (ix2 == 0) then
                    pnm0_lat2i = x2
                elseif (ix2 < -1) then
                    pnm0_lat2i = 0.0_qp
                elseif (ix2 < 0) then
                    pnm0_lat2i = x2 * bigi
                else
                    pnm0_lat2i = x2 * big
                end if
                ! .............................................................


                ! Imm
                ! .............................................................
                if (m == 1) then ! I11

                    imm0 = sqrt(3.0_qp) / 2.0_qp * &
                          ((t2 * u2 - (pi_qp / 2.0_qp - lat2i)) - &
                           (t1 * u1 - (pi_qp / 2.0_qp - lat1i)))
                    inm0 = imm0

                elseif (m == 2) then ! I22

                    imm1 = sqrt(15.0_qp) / 6.0_qp * (t2 * (3.0_qp - t2 * t2) -&
                                                     t1 * (3.0_qp - t1 * t1))
                    inm0 = imm1

                else ! I33, I44, ..., Inmax,nmax
                    
                    imm2 = gn(m + 1) * imm0 + 1.0_qp / real(m + 1, qp) * &
                           (t2 * pnm0_lat2i - t1 * pnm0_lat1i)
                    inm0 = imm2
    
                    imm0 = imm1
                    imm1 = imm2

                end if
                ! .............................................................


                ! Lumped coefficients
                ! .............................................................
                a  = inm0 * cs(m + 1, m + 1)
                b  = inm0 * cs(m, m + 1)

                if (symmi .eqv. .true.) then
                    a2 = a
                    b2 = b
                end if
                ! .............................................................
                ! -------------------------------------------------------------






                ! Tesseral Legendre functions and their integrals
                ! -------------------------------------------------------------
                if (m < nmax) then
                    
                    ! Pm+1,m for lat1i
                    ! .........................................................
                     y1 =  x1
                    iy1 = ix1
                     x1 = (am(m + 1) * t1) * y1
                    ix1 = iy1
    
                    w = abs(x1)
                    if (w >= bigs) then
                         x1 =  x1 * bigi
                        ix1 = ix1 + 1
                    elseif (w < bigsi) then
                         x1 =  x1 * big
                        ix1 = ix1 - 1
                    end if
    
                    if (ix1 == 0) then
                        pnm1_lat1i = x1
                    elseif (ix1 < - 1) then
                        pnm1_lat1i = 0.0_qp
                    elseif (ix1 < 0) then
                        pnm1_lat1i = x1 * bigi
                    else
                        pnm1_lat1i = x1 * big
                    end if
                    ! .........................................................



                    ! Pm+1,m for lat2i
                    ! .........................................................
                     y2 =  x2
                    iy2 = ix2
                     x2 = (am(m + 1) * t2) * y2
                    ix2 = iy2
    
                    w = abs(x2)
                    if (w >= bigs) then
                         x2 =  x2 * bigi
                        ix2 = ix2 + 1
                    elseif (w < bigsi) then
                         x2 =  x2 * big
                        ix2 = ix2 - 1
                    end if
    
                    if (ix2 == 0) then
                        pnm1_lat2i = x2
                    elseif (ix2 < - 1) then
                        pnm1_lat2i = 0.0_qp
                    elseif (ix2 < 0) then
                        pnm1_lat2i = x2 * bigi
                    else
                        pnm1_lat2i = x2 * big
                    end if
                    ! .........................................................


                    ! Im+1,m
                    ! .........................................................
                    ! This is not a typo, "pnm0" are indeed required here
                    inm1 = -am(m + 1) / real(m + 2, qp) * &
                            (u2 * u2 * pnm0_lat2i - &
                             u1 * u1 * pnm0_lat1i)    
                    ! .........................................................



                    ! Lumped coefficients
                    ! .........................................................
                    inmcsa = inm1 * cs(m + 2, m + 1)
                    inmcsb = inm1 * cs(m, m + 2)

                    a = a + inmcsa
                    b = b + inmcsb

                    if (symmi .eqv. .true.) then
                        a2 = a2 - inmcsa
                        b2 = b2 - inmcsb
                    end if
                    ! .........................................................


                    ! Pm+2,m, Pm+3,m, ..., Pnmax,m and their integrals
                    ! .........................................................
                    do n = (m + 2), nmax

                        ! Pm+2,m, Pm+3,m, ..., Pnmax,m for lat1i
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                        ixy1 = ix1 - iy1
    
                        if (ixy1 == 0) then
                             z1 = (am(n) * t1) * x1 - bm(n) * y1
                            iz1 = ix1
                        elseif (ixy1 == 1) then
                             z1 = (am(n) * t1) * x1 - bm(n) * (y1 * bigi)
                            iz1 = ix1
                        elseif (ixy1 == -1) then
                             z1 = (am(n) * t1) * (x1 * bigi) - bm(n) * y1
                            iz1 = iy1
                        elseif (ixy1 > 1) then
                             z1 = (am(n) * t1) * x1
                            iz1 = ix1
                        else
                             z1 = -bm(n) * y1
                            iz1 = iy1
                        end if
    
                        w = abs(z1)
                        if (w >= bigs) then
                             z1 =  z1 * bigi
                            iz1 = iz1 + 1
                        elseif (w < bigsi) then
                             z1 =  z1 * big
                            iz1 = iz1 - 1
                        end if
    
                        ! Pm+2,m, Pm+3,m, ..., Pnmax,m
                        if (iz1 == 0) then
                            pnm2_lat1i = z1
                        elseif (iz1 < -1) then
                            pnm2_lat1i = 0.0_qp
                        elseif (iz1 < 0 ) then
                            pnm2_lat1i = z1 * bigi
                        else
                            pnm2_lat1i = z1 * big
                        end if
    
                         y1 =  x1
                        iy1 = ix1
                         x1 =  z1
                        ix1 = iz1
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


                        ! Pm+2,m, Pm+3,m, ..., Pnmax,m for lat2i
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                        ixy2 = ix2 - iy2
    
                        if (ixy2 == 0) then
                             z2 = (am(n) * t2) * x2 - bm(n) * y2
                            iz2 = ix2
                        elseif (ixy2 == 1) then
                             z2 = (am(n) * t2) * x2 - bm(n) * (y2 * bigi)
                            iz2 = ix2
                        elseif (ixy2 == -1) then
                             z2 = (am(n) * t2) * (x2 * bigi) - bm(n) * y2
                            iz2 = iy2
                        elseif (ixy2 > 1) then
                             z2 = (am(n) * t2) * x2
                            iz2 = ix2
                        else
                             z2 = -bm(n) * y2
                            iz2 = iy2
                        end if
    
                        w = abs(z2)
                        if (w >= bigs) then
                             z2 =  z2 * bigi
                            iz2 = iz2 + 1
                        elseif (w < bigsi) then
                             z2 =  z2 * big
                            iz2 = iz2 - 1
                        end if
    
                        ! Pm+2,m, Pm+3,m, ..., Pnmax,m
                        if (iz2 == 0) then
                            pnm2_lat2i = z2
                        elseif (iz2 < -1) then
                            pnm2_lat2i = 0.0_qp
                        elseif (iz2 < 0 ) then
                            pnm2_lat2i = z2 * bigi
                        else
                            pnm2_lat2i = z2 * big
                        end if
    
                         y2 =  x2
                        iy2 = ix2
                         x2 =  z2
                        ix2 = iz2
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


                        ! Im+2,m, Im+3,m, ..., Inmax,m
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                        ! This is not a typo, "pnm1" are indeed required here
                        inm2 = hn(n + 1) * bm(n) * inm0 - &
                               am(n) / real(n + 1, qp) * &
                               (u2 * u2 * pnm1_lat2i - u1 * u1 * pnm1_lat1i)

                        inm0 = inm1
                        inm1 = inm2
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


                        ! Lumped coefficients
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                        inmcsa = inm2 * cs(n + 1, m + 1)
                        inmcsb = inm2 * cs(m, n + 1)

                        a = a + inmcsa
                        b = b + inmcsb

                        if (symmi .eqv. .true.) then

                            if (mod(n + m, 2) == 0) then
                                a2 = a2 + inmcsa
                                b2 = b2 + inmcsb
                            else
                                a2 = a2 - inmcsa
                                b2 = b2 - inmcsb
                            end if

                        end if
                        ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


                        pnm1_lat1i = pnm2_lat1i
                        pnm1_lat2i = pnm2_lat2i

                    end do
                    ! .........................................................

                end if
                ! -------------------------------------------------------------

            end if


            ! Before applying the PSLR algorithm from Balmino et al. (2012), 
            ! the lumped coefficients need to be multiplied by additional 
            ! terms, which stem from the integration in the longitudinal 
            ! direction
            ! ------------------------------------------------------------------
            if (m == 0) then
                m2sm2dl = dlon
            else
                m2 = 2.0_qp / real(m, qp)
                m2sm2dl = m2 * sin(dlon / m2) 
            end if

            a = a * m2sm2dl
            b = b * m2sm2dl ! Remember that "b = 0.0_qp" for "m == 0", so it can
                            ! safely be multiplied by "m2sm2dl = dlon"

            if (symmi .eqv. .true.) then
                a2 = a2 * m2sm2dl
                b2 = b2 * m2sm2dl ! Remember that "b = 0.0_qp" for "m == 0", so
                                  ! can safely be multiplied by "m2sm2dl = 
                                  ! dlon"
            end if
            ! ------------------------------------------------------------------
            

            ! The PSLR algorithm from Balmino et al. (2012) (see the REFERENCES
            ! section at the beginning of this subroutine)
            ! ------------------------------------------------------------------

            ! The first longitude from the "lon" vector (remember that the "lon 
            ! = (lon1 + lon2) / 2" and "dlon = lon1(j + 1) - lon1(j) = lon2(j + 
            ! 1) - lon2(j)")
            ! ..................................................................
             lontmp = real(m, qp) * lon0
            clontmp = cos(lontmp)
            slontmp = sin(lontmp)
            dm0     = a * clontmp + b * slontmp
            fi(1)   = fi(1) + dm0

            if (symmi .eqv. .true.) then
                dm02   = a2 * clontmp + b2 * slontmp
                fi2(1) = fi2(1) + dm02
            end if
            ! ..................................................................
            

            ! The second longitude from the "lon" vector
            ! ..................................................................
            if (nlon > 1) then

                 lontmp = real(m, qp) * (lon0 + dlon)
                clontmp = cos(lontmp)
                slontmp = sin(lontmp)
                dm1     = a * clontmp + b * slontmp
                fi(2)   = fi(2) + dm1

                if (symmi .eqv. .true.) then

                    dm12   = a2 * clontmp + b2 * slontmp
                    fi2(2) = fi2(2) + dm12
                end if

            end if
            ! ..................................................................

            
            ! The third and all the remaining longitudes from the "lon" vector
            ! ..................................................................
            cmdlon2 = 2.0_qp * cos(real(m, qp) * dlon)
            do j = 3, nlon
                
                dm2   = cmdlon2 * dm1 - dm0
                fi(j) = fi(j) + dm2
                dm0   = dm1
                dm1   = dm2

                if (symmi .eqv. .true.) then

                    dm22   = cmdlon2 * dm12 - dm02
                    fi2(j) = fi2(j) + dm22
                    dm02   = dm12
                    dm12   = dm22

                end if

            end do
            ! ..................................................................
            ! ------------------------------------------------------------------
            ! End of the PSLR algorithm
            
        end do ! End of the loop over harmonic orders
        ! ----------------------------------------------------------------------
              
         
        ! Synthesis
        ! ------------------------------------------------------------------
        ! Area of the cell on the unit sphere
        dsigma = (sin(lat2i) - sin(lat1i)) * dlon

        f(i, :) = f(i, :) + fi / dsigma
 
        if (symmi .eqv. .true.) then ! The other hemisphere computed using the
                             ! symmetry property of Legendre functions
                             ! with respect to the equator
            
            f(nlat - i + 1, :) = f(nlat - i + 1, :) + fi2 / dsigma
            
        end if
        ! ------------------------------------------------------------------

    end do ! End of the loop over evaluation points
    !$OMP END PARALLEL DO
    ! --------------------------------------------------------------------------
    ! ===============================================

end subroutine Surface_SHS_mean_qp

