subroutine Surface_SHS_qp(lat, nlat, lon, nlon, nmax, cs, ncs, f)
!
! ==========================================================================
!
! DESCRIPTION: This subroutine performs surface spherical harmonic synthesis
!              at a grid defined by the vectors of latitudes and longitudes via
!              a FFT-based algorithm (Sneeuw, 1994).
!
!              The routine is written in quadruple precision.
!
!              The loop over latitudes is parallelized using OpenMP.
!
!              The subroutine automatically starts the synthesis from degree
!              "nmin = 0". If you wish to use some other value of "nmin"
!              (but satisfying "1 <= nmin <= nmax"), you can set (outside this
!              subroutine) the respective coefficients of the "cs" matrix to
!              zero via
!
!              cs(1:nmin, 1:nmin) = 0.0_qp
!
!              where "nmin" is an integer satisfying "1 <= nmin <= nmax", and
!              then use the new matrix as the input to this subroutine. In that
!              case, the synthesis still starts at degree 0, but since the
!              coefficients up to degree "nmin - 1" are zero, they do
!              not contribute to the output signal as required.
!
!              The 4*pi fully-normalized associated Legendre functions
!              (e.g., Heiskannen and Moritz, 1967) are evaluated by the
!              numerically stable algorithm by Fukushima (2012). The algorithm
!              can be used up to ultra-high harmonic degrees (say, tens of
!              thousands or even well beyond).
!
!              The subroutine exploits the symmetry property of Legendre
!              functions with respect to the equator, thereby improving its
!              performance, especially for ultra-high-degree harmonic
!              synthesis. This computational acceleration is employed 
!              automatically, provided that the grid is recognized as symmetric 
!              with respect to the equator. Below, we provide some examples of 
!              symmetric and non-symmetric grids. Note that while these 
!              examples use the unit degrees for a more intuitive 
!              understanding, the subroutine requires radians as the input 
!              unit.
!
!              Examples of some symmetric grids (shown are only the latitudes):
!              lat = [-90, -60, -30, 0, 30, 60, 90]          ! Equator included
!              lat = [-80, -60, -40, -20, 0, 20, 40, 60, 80] ! Equator included
!              lat = [-35, -25, -15, -5, 5, 15, 25, 35]      ! Equator excluded
!              lat = [-90, -85, -80, 80, 85, 90]
!              lat = [-90, -80, -75, -70, -69, 69, 70, 75, 80, 90] ! Varying
!                                                           ! spacing, equator
!                                                           ! excluded
!
!              Examples of some grids that are not considered as symmetric
!              (shown are again only latitudes):
!              lat = [-90, -60, -30, 0, 30, 60] ! The negative latitude of -90 
!                                          ! deg does not have its positive
!                                          ! counterpart
!              lat = [-35, -25, -15, -4.99, 5, 15, 25, 35] ! The difference
!                                                   ! "abs(abs(-4.99)-5)" is
!                                                   ! larger than the threshold
!                                                   ! "thold" (see below).
!
!               In other words, the grid is symmetric, provided that all
!               positive latitudes have their negative counterpart or vice versa
!               (up to a given threshold to suppress numerical inaccuracies, see
!               the "thold" variable). The zero latitude, i.e., the equator, may
!               be included in the grid (again, within specified numerical
!               threshold "thold").
!
!
!               The longitudes of the grid must be sampled with equal spacing,
!               that is, the difference between any two consecutive longitudes
!               must be constant. The subroutine performs a check on this.
!               If the longitudes does not pass the test, a message is printed
!               and the execution is stopped. If you know for sure that
!               your longitudes share an equal spacing, but never pass
!               the test, you may want to use a larger value of the "thold"
!               variable (see below). This might happen due to the rounding
!               errors with a long vector of longitudes.
!
!
!
! INPUTS: "lat"       -- Column vector defining latitudes of the grid
!                        in radians (latitudes over a single meridian).
!
!         "nlat"      -- Number of latitudes in the array "lat".
!
!         "lon"       -- Column vector defining longitudes of the grid
!                        in radians (longitudes over a single latitude
!                        parallel).
!
!         "nlon"      -- Number of longitudes in the array "lon".
!
!         "nmax"      -- Maximum spherical harmonic degree of the synthesis
!                        (Non-negative integer, 0 <= nmax < ncs). See below
!                        for the explanation of the "ncs" variable.
!
!         "cs"        -- Matrix of dimensions (ncs, ncs) with
!                        input spherical harmonic coefficients (see below
!                        the explanation of "ncs"). The structure
!                        of the matrix is as follows:
!
!                        [C(0, 0) S(1, 1) S(2, 1)   ...          S(ncs - 1, 1)]
!                        [C(1, 0) C(1, 1) S(2, 2)   ...          S(ncs - 1, 2)]
!                        [C(2, 0) C(2, 1) C(2, 2)   ...          S(ncs - 1, 3)]
!                        [  .       .      .        .                  .      ]
!                        [  .       .      .         .                 .      ]
!                        [  .       .      .          .                .      ]
!                        [C(ncs - 1, 0)             ...    C(ncs - 1, ncs - 1)]
!
!                        where C(n, m) is the coefficient "C" of degree "n" and
!                        order "m", etc.
!
!         "ncs"       -- Number of rows of the "cs" matrix (equals to the                       
!                        number of columns of "cs")
!
!
! OUTPUTS: "f" -- Matrix of the synthesized signal (e.g., topographic heights)
!                 of dimensions (nlat, nlon) and with the following structure
!
!                 [(lat(1),lon(1))      ...      (lat(1),lon(nlon)]
!                 [      .              .               .         ]
!                 [      .               .              .         ]
!                 [      .                .             .         ]
!                 [(lat(nlat),lon(1))   ...   (lat(nlat),lon(nlon)]
!
!
! REFERENCES: Sneeuw, N. (1994) Global spherical harmonic analysis by
!                least-squares and numerical quadrature methods in historical
!                perspective. Geophysical Journal International 118:707--716
!
!             Heiskannen, W. A., Moritz, H. (1967) Physical Geodesy. W. H.
!                Freeman and Company, San Francisco, 364 pp
!
!             Fukushima, T. (2012) Numerical computation of spherical
!                harmonics of arbitrary degree and order by extending exponent
!                of floating point numbers. Journal of Geodesy 86:271--285.
!
!             Balmino, Vales, Bonvalot, Briais, 2012. Spherical harmonic
!                 modelling to ultra-high degree of Bouguer and isostatic
!                 anomalies. Journal of Geodesy 86:499–520,
!                 doi: 10.1007/s00190-011-0533-4
!
!
! Contact: blazej.bucha@stuba.sk
!
!
! Please use the following reference when using this subroutine:
!
!          Bucha, B., Hirt, C., Kuhn, M., 2019. Cap integration in
!             spectral gravity forward modelling up to the full gravity
!             tensor. Journal of Geodesy 93:1707--1737
!
! =========================================================================






    ! Import modules
    ! ===============================================
    use vartypes,  only: qp ! Module defining number formats
    use constants, only: ind_qp, root3_qp, eps_qp ! Module defining constants
                                                  ! (pi, etc.)
    use FFTW3 ! Module for the FFTW package
    ! ===============================================






    ! Declaration of variables
    ! ===============================================
    implicit none

    integer,  intent(in) :: nmax ! Maximum degree of SH expansion
    integer,  intent(in) :: nlat ! Number of latitudes in one meridian of
                                 ! the grid
    integer,  intent(in) :: nlon ! Number of longitudes in one parallel of the
                                 ! grid
    real(qp), intent(in) :: lat(nlat), lon(nlon)  ! Spherical latitudes and
                                                  ! longitudes of the grid
                                                  ! points in radians
    integer,  intent(in) :: ncs  ! Number of rows of the "cs" matrix (equals to
                                 ! number of columns of "cs")
    real(qp), intent(in) :: cs(ncs, ncs)  ! Matrix of spherical
                                          ! harmonic coefficients


    ! Legendre functions (LFs) and some variables necessary for their evaluation
    ! via the extended-range arithmetic
    real(qp), parameter :: big = 2.0_qp**ind_qp
    real(qp), parameter :: bigi = 2.0_qp**(-ind_qp)
    real(qp), parameter :: bigs = 2.0_qp**(ind_qp / 2)
    real(qp), parameter :: bigsi = 2.0_qp**(-ind_qp / 2)
    real(qp) :: pnm2, pnm1, pnm0   ! LFs
    real(qp) :: am(nmax + 1), bm(nmax), dn(nmax), en(nmax + 1), fn(nmax + 1)
    real(qp) :: ps(nmax), r(2 * nmax + 3), ri(2 * nmax + 3)
    real(qp) :: u, t, x, y, w, z
    integer  :: ix, iy, iz, ixy, ips(nmax)

    ! Some variables for the PSLR algorithm
    real(qp) :: dm0, dm1, dm2
    real(qp) :: dm02, dm12, dm22
    real(qp) :: lon0, dlon, cmdlon2
    real(qp) :: lontmp, clontmp, slontmp

    real(qp) :: dlonchck ! Auxiliary variable to check the longitude step
                         ! of the grid

    integer :: even    ! If the number of latitudes is even, then "even = 1",
                       ! otherwise "even = 0"
    integer :: symm    ! If the grid is symmetric with respect to the equator,
                       ! then "symm = 1", otherwise "symm = 0". If "symm == 1",
                       ! the subroutine automatically exploits the symmetry
                       ! property of Legendre functions in order to accelerate
                       ! the computation
    integer :: nlatdo  ! If the grid is symmetric with respect to the equator,
                       ! then "nlatdo" represents the number of latitudes from
                       ! one of the poles to the equator (including)
    logical :: symmi   ! Just some auxiliary variable

    integer :: i, j, n, m  ! Loop variables over the latitudes, longitudes,
                           ! harmonic degrees and orders

    real(qp), parameter :: thold = 10000.0_qp * eps_qp
        ! If abs(a - b) <= thold, then "a" is considered in this subroutine to
        ! be equal to "b" within the threshold "thold". "thold" is used to check
        ! whether input the latitudes and longitudes meet some criteria
        ! (symmetric latitudes with respect to the equator, etc.).
        ! The value of "thold" can be modified by the user if necessary.
        ! The larger the value, the larger numerical inaccuracies will be
        ! tolerated.

    real(qp) :: a, b, a2, b2        ! Lumped coefficients
    real(qp) :: pnmcsa, pnmcsb      ! Legendre functions times spherical
                                    ! harmonic coefficients to compute lumped
                                    ! coefficients
    real(qp) :: fi(nlon), fi2(nlon) ! Temporary contribution of the ith
                                    ! latitude parallel to the output quantity

    real(qp), intent(out) :: f(nlat, nlon) ! Output synthesized quantity
    ! ===============================================







    ! The main computational part of the subroutine
    ! ===============================================

    ! Check the latitudes of the computational grid
    ! --------------------------------------------------------------------------
    ! Check whether the number of latitudes is even or odd
    ! ..........................................................................
    if (mod(nlat, 2) .eq. 0) then
        even = 1  ! The number of latitudes is an even number and the grid does
                  ! not contain the zero latitude.
    else
        even = 0  ! The number of latitudes is an odd number and the grid may
                  ! contain the zero latitude.
    end if
    ! ..........................................................................


    ! Determine whether the latitudes are symmetric with respect to the equator
    ! ..........................................................................
    if (nlat == 1) then ! If there is only one latitude within the grid,
                        ! the grid is automatically considered as not symmetric
        symm = 0
    else
        symm = 1  ! If there is more than one latitude in the grid, let's start
                  ! by assuming that the grid is symmetric with respect to the
                  ! equator and check whether this is indeed true
    end if

    do i = 1, ((nlat + even - 1) / 2)
        if (abs(lat(i) + lat(nlat - i + 1)) > thold) then

            symm = 0  ! The grid is not symmetric.

            exit  ! Exiting the loop with the final decision: the grid is not
                  ! symmetric with respect to the equator ("symm = 0")
        end if
    end do


    ! If the grid is symmetric, check whether the zero latitude is present right
    ! in the middle of the "lat" vector. If "even == 1", we already know that
    ! the grid does not contain the zero latitude
    if ((symm == 1) .and. (even == 0)) then
        if (abs(lat((nlat + 1) / 2)) > thold) then ! The middle latitude
            ! is not zero, so the grid is not clasified as symmetric with
            ! respect to the equator
            symm = 0  ! The final decision: the grid is not symmetric with
                      ! respect to the equator ("symm = 0")
        end if
    end if
    ! ..........................................................................


    ! Finally, if the grid is symmetric, we modify the number of latitudes
    ! "nlat" to be equal to the number of latitudes on one hemisphere only
    ! (including the equator if present). This is because the
    ! time-consuming "do loop" over evaluation points now needs to run for one
    ! hemisphere only, while the results for the other hemishere are obtained
    ! by exploiting the symmetry property of Legendre functions. This reduces
    ! the number of Legendre functions that need to be evaluated by a factor of
    ! ~2, so saves some computational time
    ! ..........................................................................
    if (symm == 1) then
        if (even == 1) then
            nlatdo = nlat / 2
        else
            nlatdo = (nlat + 1) / 2
        end if
    else
        nlatdo = nlat
    end if
    ! ..........................................................................
    ! --------------------------------------------------------------------------






    ! Check the longitudes of the computational grid
    ! --------------------------------------------------------------------------
    ! Get the origin of the longitude "lon" vector (will be necessary later for
    ! the PSLR algorithm)
    lon0 = lon(1)


    ! Check whether the first element of "lon" is the smallest one. Otherwise,
    ! the execution of the code is terminated.
    if (abs(lon0 - minval(lon)) > thold) then
        error stop '[Surface_SHS_qp.f95 says:] The first element of the &
              &longitude vector "lon" has to be the smallest one. &
              &Terminating the code.'
    end if


    ! Get the lon step of the grid (will be necessary later for the PSLR
    ! algorithm)
    if (nlon > 1) then
        dlon = lon(2) - lon(1)
    else
        dlon = 0.0_qp
    end if


    ! Check whether the longitude step is constant for the entire "lon" vector.
    ! Otherwise, the execution of the code is terminated.
    do j = 2, nlon
        dlonchck = lon(j) - lon(j - 1)

        if (abs(dlon - dlonchck) > thold) then
            error stop '[Surface_SHS_qp.f95 says:] The longitudinal step of &
                  &the grid has to be constant. Terminating the code.'
        end if
    end do
    ! --------------------------------------------------------------------------






    ! Initializations for recurrence relations to compute Legendre functions
    ! --------------------------------------------------------------------------
    do m = 1, (2 * nmax + 3)
        w = sqrt(real(m, qp))
        r(m) = w
        ri(m) = 1.0_qp / w
    end do


    ! Computation of the "dn" coefficients for Legendre recurrence 
    ! relations
    do n = 1, nmax
        dn(n) = r(2 * n + 3) * ri(2 * n + 2)
    end do


    ! Computation of "en" and "fn" coefficients for Legendre recurrence
    ! relations
    do n = 2, nmax
        en(n + 1) = r(2 * n + 1) * r(2 * n - 1) / real(n, qp)
        fn(n + 1) = real(n - 1, qp) * r(2 * n + 1) / (real(n, qp) * &
                   &r(2 * n - 3))
    end do
    ! --------------------------------------------------------------------------







    ! Initialize all elements of the output signal "f" to zero
    ! --------------------------------------------------------------------------
    f = 0.0_qp
    ! --------------------------------------------------------------------------


    ! Loop over grid latitudes
    ! --------------------------------------------------------------------------
    !$OMP PARALLEL DO DEFAULT(NONE) &
    !$OMP& SHARED(f, cs, nmax, lat, dn, en, fn, r, ri, nlat, nlon, nlatdo) &
    !$OMP& SHARED(lon0, dlon, even, symm) &
    !$OMP& PRIVATE(pnm0, pnm1, pnm2, t, u, x, ix, y, iy, ixy, z, iz) &
    !$OMP& PRIVATE(ps, ips, fi, fi2, am, w, bm) &
    !$OMP& PRIVATE(a, b, a2, b2, pnmcsa, pnmcsb, lontmp) &
    !$OMP& PRIVATE(clontmp, slontmp, dm0, dm1, dm2, dm02, dm12, dm22) &
    !$OMP& PRIVATE(cmdlon2, symmi)
    do i = 1, nlatdo

        ! Check whether the symmetry property of LFs needs to be exploited
        ! ------------------------------------------------------------------
        if (nlat == 1) then ! There is only a single latitude parallel
                            ! in the grid, so no symmetry property
            symmi = .false.
        elseif (symm == 0) then ! The grid was already classified as
                                ! non-symmetric
            symmi = .false.
        elseif (symm == 1 .and. even == 0 .and. i == nlatdo) then
            symmi = .false.  ! For a symmetric grid containing the equator,
                             ! this ensures that the points on the equator
                             ! will not be processed twice
        else
            symmi = .true. ! Exploit the symmetry property
        end if
        ! ------------------------------------------------------------------


        ! Some precomputations for Legendre functions
        ! ------------------------------------------------------------------
        t  = sin(lat(i))
        u  = cos(lat(i))
        x  = root3_qp * u
        ix = 0

         ps(1) =  x
        ips(1) = ix


        do n = 2, nmax
            x = (dn(n - 1) * u) * x
            y = abs(x)

            if (y >= bigs) then
                 x =  x * bigi
                ix = ix + 1
            elseif (y < bigsi) then
                 x =  x * big
                ix = ix - 1
            end if

             ps(n) =  x
            ips(n) = ix
        end do
        ! ------------------------------------------------------------------


        ! The "fi" vector represents the synthesized quantity "f" for the ith
        ! latitude parallel. Therefore, it needs to be reinitialized to zero
        ! for each "ith" latitude. The same holds true for "fi2".
        fi  = 0.0_qp
        fi2 = 0.0_qp


        ! Loop over harmonic orders
        ! ----------------------------------------------------------------------
        do m = 0, nmax

            ! Computation of "am" and "bm" coefficients for Legendre recurrence
            ! relations
            ! ------------------------------------------------------------------
            am(m + 1) = r(2 * m + 3)
            do n = (m + 2), nmax
                w     = r(2 * n + 1) * ri(n - m) * ri(n + m)
                am(n) = r(2 * n - 1) * w
                bm(n) = r(n - m - 1) * r(n + m - 1) * ri(2 * n - 3) * w
            end do
            ! ------------------------------------------------------------------


            ! Computation of the lumped coefficients
            ! ------------------------------------------------------------------
            if (m == 0) then

                ! Zonal harmonics
                ! --------------------------------------------------------------
                ! P00
                pnm0 = 1.0_qp
                a = pnm0 * cs(1, m + 1)
                b = 0.0_qp

                if (symmi .eqv. .true.) then
                    a2 = a
                    b2 = 0.0_qp
                end if

                ! P10
                if (nmax >= 1) then

                    pnm1   = root3_qp * t
                    pnmcsa = pnm1 * cs(2, m + 1)
                    a      = a + pnmcsa

                    if (symmi .eqv. .true.) then
                        a2 = a2 - pnmcsa
                    end if

                end if

                ! P20, P30, ..., Pnmax,0
                if (nmax >= 2) then

                    do n = 2, nmax

                        pnm2   = en(n + 1) * t * pnm1 - fn(n + 1) * pnm0
                        pnmcsa = pnm2 * cs(n + 1, m + 1)
                        a      = a + pnmcsa

                        if (symmi .eqv. .true.) then
                            if (mod(n, 2) == 0) then
                                a2 = a2 + pnmcsa
                            else
                                a2 = a2 - pnmcsa
                            end if
                        end if

                        pnm0 = pnm1
                        pnm1 = pnm2

                    end do

                end if
                ! --------------------------------------------------------------


            else ! Non-zonal harmonics


                ! Sectorial harmonics
                ! --------------------------------------------------------------
                x = ps(m)
                ix = ips(m)

                ! Pmm
                if (ix == 0) then
                    pnm0 = x
                elseif (ix < -1) then
                    pnm0 = 0.0_qp
                elseif (ix < 0) then
                    pnm0 = x * bigi
                else
                    pnm0 = x * big
                end if

                a  = pnm0 * cs(m + 1, m + 1)
                b  = pnm0 * cs(m, m + 1)

                if (symmi .eqv. .true.) then
                    a2 = a
                    b2 = b
                end if
                ! --------------------------------------------------------------


                ! Tesseral harmonics
                ! --------------------------------------------------------------
                if (m < nmax) then
                     y =  x
                    iy = ix
                     x = (am(m + 1) * t) * y
                    ix = iy

                    w = abs(x)
                    if (w >= bigs) then
                         x =  x * bigi
                        ix = ix + 1
                    elseif (w < bigsi) then
                         x =  x * big
                        ix = ix - 1
                    end if

                    ! Pm+1,m
                    if (ix == 0) then
                        pnm1 = x
                    elseif (ix < - 1) then
                        pnm1 = 0.0_qp
                    elseif (ix < 0) then
                        pnm1 = x * bigi
                    else
                        pnm1 = x * big
                    end if

                    pnmcsa = pnm1 * cs(m + 2, m + 1)
                    pnmcsb = pnm1 * cs(m, m + 2)

                    a = a + pnmcsa
                    b = b + pnmcsb

                    if (symmi .eqv. .true.) then
                        a2 = a2 - pnmcsa
                        b2 = b2 - pnmcsb
                    end if

                    ! Loop over degrees
                    ! ----------------------------------------------------------
                    do n = (m + 2), nmax

                        ixy = ix - iy

                        if (ixy == 0) then
                             z = (am(n) * t) * x - bm(n) * y
                            iz = ix
                        elseif (ixy == 1) then
                             z = (am(n) * t) * x - bm(n) * (y * bigi)
                            iz = ix
                        elseif (ixy == -1) then
                             z = (am(n) * t) * (x * bigi) - bm(n) * y
                            iz = iy
                        elseif (ixy > 1) then
                             z = (am(n) * t) * x
                            iz = ix
                        else
                             z = -bm(n) * y
                            iz = iy
                        end if

                        w = abs(z)
                        if (w >= bigs) then
                             z =  z * bigi
                            iz = iz + 1
                        elseif (w < bigsi) then
                             z =  z * big
                            iz = iz - 1
                        end if

                        ! Pm+2,m, Pm+3,m, ..., Pnmax,m
                        if (iz == 0) then
                            pnm2 = z
                        elseif (iz < -1) then
                            pnm2 = 0.0_qp
                        elseif (iz < 0 ) then
                            pnm2 = z * bigi
                        else
                            pnm2 = z * big
                        end if

                        pnmcsa = pnm2 * cs(n + 1, m + 1)
                        pnmcsb = pnm2 * cs(m, n + 1)

                        a = a + pnmcsa
                        b = b + pnmcsb

                        if (symmi .eqv. .true.) then

                            if (mod(n + m, 2) == 0) then
                                a2 = a2 + pnmcsa
                                b2 = b2 + pnmcsb
                            else
                                a2 = a2 - pnmcsa
                                b2 = b2 - pnmcsb
                            end if

                        end if

                         y =  x
                        iy = ix
                         x =  z
                        ix = iz

                    end do ! End of the loop over harmonic degrees
                    ! ----------------------------------------------------------


                end if ! End of computation of tesseral harmonics
                ! --------------------------------------------------------------


            end if ! End of computation of the lumped coefficients


            ! The PSLR algorithm from Balmino et al. (2012) (see the REFERENCES
            ! section at the beginning of this subroutine)
            ! ------------------------------------------------------------------

            ! The first longitude from the "lon" vector
            ! ..................................................................
             lontmp = real(m, qp) * lon0
            clontmp = cos(lontmp)
            slontmp = sin(lontmp)
            dm0     = a * clontmp + b * slontmp
            fi(1)   = fi(1) + dm0

            if (symmi .eqv. .true.) then
                dm02   = a2 * clontmp + b2 * slontmp
                fi2(1) = fi2(1) + dm02
            end if
            ! ..................................................................


            ! The second longitude from the "lon" vector
            ! ..................................................................
            if (nlon > 1) then

                 lontmp = real(m, qp) * (lon0 + dlon)
                clontmp = cos(lontmp)
                slontmp = sin(lontmp)
                dm1     = a * clontmp + b * slontmp
                fi(2)   = fi(2) + dm1

                if (symmi .eqv. .true.) then

                    dm12   = a2 * clontmp + b2 * slontmp
                    fi2(2) = fi2(2) + dm12
                end if

            end if
            ! ..................................................................


            ! The third and all the remaining longitudes from the "lon" vector
            ! ..................................................................
            cmdlon2 = 2.0_qp * cos(real(m, qp) * dlon)
            do j = 3, nlon

                dm2   = cmdlon2 * dm1 - dm0
                fi(j) = fi(j) + dm2
                dm0   = dm1
                dm1   = dm2

                if (symmi .eqv. .true.) then

                    dm22   = cmdlon2 * dm12 - dm02
                    fi2(j) = fi2(j) + dm22
                    dm02   = dm12
                    dm12   = dm22

                end if

            end do
            ! ..................................................................
            ! ------------------------------------------------------------------
            ! End of the PSLR algorithm

        end do ! End of the loop over harmonic orders
        ! ----------------------------------------------------------------------


        ! Synthesis
        ! ------------------------------------------------------------------
        f(i, :) = f(i, :) + fi

        if (symmi .eqv. .true.) then ! The other hemisphere computed using the
                             ! symmetry property of Legendre functions
                             ! with respect to the equator

            f(nlat - i + 1, :) = f(nlat - i + 1, :) + fi2

        end if
        ! ------------------------------------------------------------------

    end do ! End of the loop over latitude parallels
    !$OMP END PARALLEL DO
    ! --------------------------------------------------------------------------
    ! ===============================================

end subroutine Surface_SHS_qp
