function [degree,order,C,S]=Surface_SHA(lat,w,signal,nmax_signal,nmax_SHA)
%==========================================================================
%DESCRIPTION: This function computes spherical harmonic coefficients of a
%             signal via the exact FFT-based Gauss--Legendre quadrature
%             (Sneeuw, 1994). The signal must be sampled at the nodes of 
%             the Gauss--Legendre quadrature (see the "GL_grid" function)
%             and possess harmonics up to degree "nmax_signal" (see below).
%
%             The 4*pi fully-normalized associated Legendre functions
%             (e.g., Heiskannen and Moritz, 1967) are evaluated by the 
%             numerically stable algorithm by Fukushima (2012). The algorithm 
%             can be used up to ultra-high harmonic degrees (say, tens of
%             thousands or even well beyond).
%
%             As the main input, the signal to be analyzed is represented
%             by a matrix as one of the input variables (see "signal" below).
%             Importantly, the whole matrix is stored in RAM during the
%             entire harmonic analysis. The same holds for the output spherical
%             harmonic coefficients. Ultra-high-degree analysis may therefore
%             require tens of GBs of RAM. As a possible solutions, the 
%             enclosed function "Surface_SHA_sliced" is design to significantly 
%             reduce the RAM requirements.
%
%             The function exploits the symmetry propery of Legendre functions 
%             with respect to the equator, thereby improving its performance,
%             especially for ultra-high-degree harmonic analysis.
%
%INPUTS: "lat"         -- Latitudes of the Gauss--Legendre grid in radians. 
%                         A vector of the dimension (nmax+1,1). See
%                         the "GL_grid.m" function.
%
%        "w"           -- Weights given by the Gauss--Legendre quadrature.
%                         A vector of the dimension (nmax+1,1). See
%                         the "GL_grid.m" function.
%
%        "signal"      -- Input signal i) given at the nodes of the Gauss--Legendre
%                         quadrature and ii) containing harmonics up to degree
%                         "nmax_signal" (see below).
%                         A matrix of the dimensions (nmax+1,2*nmax+2),
%                         sampled as follows:
%
%                         signal = [(lat(1),lon(1))              ...                (lat(1),lon(2*nmax_signal+2)]
%                                  [      .                      .                               .              ]
%                                  [      .                       .                              .              ]
%                                  [      .                        .                             .              ]
%                                  [(lat(nmax_signal+1),lon(1))   ...   (lat(nmax_signal+1),lon(2*nmax_signal+2)]
%
%                         where "lat" and "lon" are vectors generated by
%                         the function "GL_grid" using the following command
%
%                         "[lat,lon,w]=GL_grid(nmax_signal);"
%
%                         See also the "Surface_SHS.m" function.
%
%        "nmax_signal" -- Maximum harmonic degree present in the signal
%                         (here, the variable "signal" above).
%
%        "nmax_SHA"    -- Maximum harmonic degree of the output spherical
%                         harmonic coefficients. This value cannot be larger
%                         than "nmax_signal". Note that if
%                         "nmax_SHA<nmax_signal", the output coefficients
%                         are still computed exactly, because the matrix
%                         "signal" is sampled correctly according to the
%                         maximum degree that is present in the signal, here
%                         "nmax_signal".
%
%OUTPUTS: "degree" -- Vector indicating harmonic degree of respective 
%                     spherical harmonic coefficients C and S.
%                     Dimension (sum(0:(nmax_SHA+1)),1).
%         "order"  -- Vector indicating harmonic order of respective
%                     spherical harmonic coefficients C and S.
%                     Dimension (sum(0:(nmax_SHA+1)),1).
%         "C"      -- Vector of spherical harmonic coefficients C.
%                     Dimension (sum(0:(nmax_SHA+1)),1).
%         "S"      -- Vector of spherical harmonic coefficients S.
%                     Dimension (sum(0:(nmax_SHA+1)),1).
%
%REFERENCES: Sneeuw, N. (1994) Global spherical harmonic analysis by
%               least-squares and numerical quadrature methods in historical perspective.
%               Geophysical Journal International 118:707--716
%
%            Heiskannen, W. A., Moritz, H. (1967) Physical Geodesy. W. H.
%               Freeman and Company, San Francisco, 364 pp
%
%            Fukushima, T. (2012) Numerical computation of spherical
%               harmonics of arbitrary degree and order by extending exponent
%               of floating point numbers. Journal of Geodesy 86:271--285.
%
%Contact: blazej.bucha@stuba.sk
%
%
%Please use the following reference when using this function:
%
%         Bucha, B., Hirt, C., Kuhn, M., 2019. Cap integration in
%            spectral gravity forward modelling up to the full gravity
%            tensor. Journal of Geodesy 93:1707--1737
%
%==========================================================================



%Initialization of matrix with spherical harmonic coefficients
Cnm=zeros(nmax_SHA+1); %The structure of the matrix Cnm is as follows:
                       %
                       %[C(0,0) S(1,1) S(2,1)    ...           S(nmax_SHA,1)]
                       %[C(1,0) C(1,1) S(2,2)    ...           S(nmax_SHA,2)]
                       %[C(2,0) C(2,1) C(2,2)    ...           S(nmax_SHA,3)]
                       %[  .       .      .      .                   .      ]
                       %[  .       .      .       .                  .      ]
                       %[  .       .      .        .                 .      ]
                       %[C(nmax_SHA,0)   ...            C(nmax_SHA,nmax_SHA)]
                       %
                       %At the end of the code, the matrix Cnm is
                       %transformed into the output variables 
                       %[degree order C S].

                       
%Identification of the grid-type to utilize the symmetry property of
%Legendre functions
%==========================================================================                      
length_lat=length(lat); %The total number of latitudes in the grid
Pnm_all=zeros(1,length_lat);
floor_length_lat2=floor(length_lat/2);
if rem(nmax_signal,2)==0 %If the Gauss--Legendre grid contains the zero latitude
    lat=lat(1:floor_length_lat2+1); %Latitudes from the interval (-90 deg, 0 deg],
                                    %(including the zero latitude)
    even=1;
else
    lat=lat(1:floor_length_lat2); %Latitudes from the interval (-90 deg, 0 deg),
                                  %(the grid does not contain the zero
                                  %latitude)
    even=0;
end
length_lat=length(lat); %The length of the NEW vector "lat" with the 
                        %non-positive latitudes
%==========================================================================
                        

%Lumped coefficents
%==========================================================================
FFT=fft(signal,[],2);
clear signal

A=bsxfun(@times,real(FFT(:,1:nmax_signal+1)),2*w*pi/(2*(nmax_signal+1)));
B=bsxfun(@times,-imag(FFT(:,1:nmax_signal+1)),2*w*pi/(2*(nmax_signal+1)));

clear w FFT
%==========================================================================


%Initializations for the computation of Legendre functions
%==========================================================================
nmax23=nmax_SHA*2+3;
rr=zeros(nmax23,1); ri=rr;
dd=zeros(nmax_SHA,1); am=dd; bm=am;

m1=1:nmax23;
rr(m1)=sqrt(m1);
ri(m1)=1./rr;
m2=1:nmax_SHA;
dd(m2)=rr(2*m2+3).*ri(2*m2+2);

IND=960;
BIG=2^IND;
BIGI=2^(-IND);
BIGS=2^(IND/2);
BIGSI=2^(-IND/2);
ROOT3=1.732050807568877;

u(1:length_lat)=cos(lat);
t(1:length_lat)=sin(lat);
clear lat
temp1=zeros(1,length_lat);
temp2=ones(1,length_lat);
temp3=temp2;
temp4=temp1;
temp5=temp1+BIGI;
ps1b=zeros(nmax_SHA,length_lat); 
ips1b=ps1b;
xb=ROOT3*u;
ixb=zeros(size(xb));
ps1b(1,:)=xb;
ips1b(1,:)=ixb;
for m3=2:nmax_SHA
    xb=(dd(m3-1)*u).*xb;
    yb=abs(xb);
    iyb=yb>=BIGS;
    if any(iyb)
        xb(iyb)=xb(iyb)*BIGI;
        ixb(iyb)=ixb(iyb)+1;
    end
    iyb=yb<BIGSI;
    if any(iyb)
        xb(iyb)=xb(iyb)*BIG;
        ixb(iyb)=ixb(iyb)-1;
    end
    ps1b(m3,:)=xb;
    ips1b(m3,:)=ixb;
end
%==========================================================================


%Loop over spherical harmonic order
%==========================================================================
for m=0:nmax_SHA
    
    am(m+1)=rr(2*m+3);
    for n=m+2:nmax_SHA
        ww=rr(2*n+1)*ri(n-m)*ri(n+m);
        am(n)=rr(2*n-1)*ww;
        bm(n)=rr(n-m-1)*rr(n+m-1)*ri(2*n-3)*ww;
    end
    
    if m==0 %Zonal spherical harmonic coefficients

        Pnm0=ones(1,length_lat);
        
        Pnm_all(1:length_lat)=Pnm0;
        Pnm_all((length_lat+1):end)=Pnm0((end-even):-1:1);
        
        Cnm(1,m+1)=Pnm_all*A(:,m+1);
        
        if nmax_SHA>0
            Pnm1=sqrt(3)*t;
            
            Pnm_all(1:length_lat)=Pnm1;
            Pnm_all((length_lat+1):end)=-Pnm1((end-even):-1:1);
            
            Cnm(2,m+1)=Pnm_all*A(:,m+1);
        end
        
        if nmax_SHA>1              
            for i=2:nmax_SHA
                Pnm2=Pnm1.*sqrt((2*i+1)*(2*i-1))./i.*t-Pnm0.*(i-1).*sqrt(2.*i+1)./(i.*sqrt(2.*i-3));
                
                Pnm_all(1:length_lat)=Pnm2;
                if rem(i,2)==0
                    Pnm_all((length_lat+1):end)=Pnm2((end-even):-1:1);
                elseif rem(i,2)==1
                    Pnm_all((length_lat+1):end)=-Pnm2((end-even):-1:1);
                end
                
                Cnm(i+1,m+1)=Pnm_all*A(:,m+1);

                Pnm0=Pnm1;
                Pnm1=Pnm2;
            end
        end
        
    elseif m~=0 %Non-zonal spherical harmonic coefficients                              
        xb=ps1b(m,:);
        ixb=ips1b(m,:);
                                           
        temp5(ixb==0)=1;
        temp5(ixb<-1)=0;
        %temp5(izb>=-1 & izb<0)=BIGI;
        %The condition "izb>=-1 & izb<0"
        %is useless, as "izb" is already
        %initialized as "izb=BIGI".
        temp5(ixb>0)=BIG;
        
        Pnm0=xb.*temp5;
        temp5=temp5.*0+BIGI; 
        
        Pnm_all(1:length_lat)=Pnm0;
        Pnm_all((length_lat+1):end)=Pnm0((end-even):-1:1);
        
        %Sectorial spherical harmonic coefficients
        Cnm(m+1,m+1)=Pnm_all*A(:,m+1);
        Cnm(m,m+1)=Pnm_all*B(:,m+1);
        
        if m<nmax_SHA 
            yb=xb;
            iyb=ixb;

            xb=(am(m+1).*t).*yb;
            ixb=iyb;
            wb=abs(xb);

            wBIGSb=wb>=BIGS;
            wBIGSIb=wb<BIGSI;
            temp3(wBIGSb)=BIGI;
            temp3(wBIGSIb)=BIG;
            temp4(wBIGSb)=1;
            temp4(wBIGSIb)=-1;

            xb=xb.*temp3;
            ixb=ixb+temp4;
            temp3=temp2;
            temp4=temp4.*0;
           
            temp5(ixb==0)=1;
            temp5(ixb<-1)=0;
            %temp5(izb>=-1 & izb<0)=BIGI;
            %The condition "izb>=-1 & izb<0"
            %is useless, as "izb" is already
            %initialized as "izb=BIGI".
            temp5(ixb>0)=BIG;
               
            Pnm1=xb.*temp5;
            temp5=temp5.*0+BIGI; 

            Pnm_all(1:length_lat)=Pnm1;
            Pnm_all((length_lat+1):end)=-Pnm1((end-even):-1:1);
            
            %Tesseral spherical harmonic coefficients
            Cnm(m+2,m+1)=Pnm_all*A(:,m+1);
            Cnm(m,m+2)=Pnm_all*B(:,m+1);
            
            if m<nmax_SHA-1
                for n=m+2:nmax_SHA
                    idb=ixb-iyb;

                    id0b=idb==0;
                    id1b=idb==1;
                    id_1b=idb==-1;
                    idv1b=idb>1;

                    temp1(id0b)=1;
                    temp1(id1b)=1;
                    temp2(id1b)=BIGI;
                    temp1(id_1b)=BIGI;
                    temp1(idv1b)=1;
                    temp2(idv1b)=0;

                    zzb=(am(n).*t).*(xb.*temp1)-bm(n).*(yb.*temp2);
                    izb=iyb;
                    id0b_id1b_idv1b=id0b | id1b | idv1b;
                    izb(id0b_id1b_idv1b)=ixb(id0b_id1b_idv1b);
                    temp1=temp1.*0;
                    temp2=temp1+1;

                    wb=abs(zzb);

                    wBIGSb=wb>=BIGS;
                    wBIGSIb=wb<BIGSI;
                    temp3(wBIGSb)=BIGI;
                    temp3(wBIGSIb)=BIG;
                    temp4(wBIGSb)=1;
                    temp4(wBIGSIb)=-1;

                    zzb=zzb.*temp3;
                    izb=izb+temp4;
                    temp3=temp2;
                    temp4=temp4.*0;

                    temp5(izb==0)=1;
                    temp5(izb<-1)=0;
                    %temp5(izb>=-1 & izb<0)=BIGI;
                    %The condition "izb>=-1 & izb<0"
                    %is useless, as "izb" is already
                    %initialized as "izb=BIGI".
                    temp5(izb>0)=BIG;
                    
                    Pnm=zzb.*temp5;
                    temp5=temp1+BIGI;  
                    
                    %Tesseral spherical harmonic coefficients
                    Pnm_all(1:length_lat)=Pnm;
                    if rem(n+m,2)==0
                        Pnm_all((length_lat+1):end)=Pnm((end-even):-1:1);
                    elseif rem(n+m,2)==1
                        Pnm_all((length_lat+1):end)=-Pnm((end-even):-1:1);
                    end
                    
                    Cnm(n+1,m+1)=Pnm_all*A(:,m+1);
                    Cnm(m,n+1)=Pnm_all*B(:,m+1);
                    
                     

                    yb=xb;
                    iyb=ixb;
                    xb=zzb;
                    ixb=izb;
                end
            end
        end
    end
end
clear A B Pnm Pnm0 Pnm1 Pnm2 am bm dd fi id0b id0b_id1b_idv1b id1b idb idv1b id_1b id_1b...
    idb1b ips1b ixb iyb izb m1 m2 ps1b ri rr t temp1 temp2 temp3 temp4 temp5...
    u wBIGSIb wBIGSb wb xb yb zzb

Cnm=Cnm./(4*pi);
%==========================================================================


%Transformation of the matrix Cnm into the standard format as used, for
%instance, with spherical harmonic gravity field models (see, e.g., 
%the structure of gfc files used by ICGEM at http://icgem.gfz-potsdam.de)
%==========================================================================
degree=zeros(sum(0:(nmax_SHA+1)),1);
order=degree;
C=degree;
S=degree;
rows=nmax_SHA+1;
first=1;
last=nmax_SHA+1;
for i=1:rows   
    degree(first:last)=[(i-1):(rows-1)]'; %#ok<NBRAK>
    order(first:last)=(i-1)*ones(length(Cnm(i:end,i)),1);

    C(first:last)=Cnm(i:end,i);
    if i==1
        S(first:last)=0;
    else
        S(first:last)=Cnm(i-1,i:end);
    end

    first=last+1;
    last=last+1+(nmax_SHA-i);
end
%==========================================================================
